# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import math
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Dict, List, Set, Tuple, TypeVar, Union

from rai.types import Pose
from rclpy.impl.rcutils_logger import RcutilsLogger

from rai_sim.simulation_bridge import (
    Entity,
    SimulationBridge,
    SimulationConfig,
    SimulationConfigT,
    SpawnedEntity,
)

loggers_type = Union[RcutilsLogger, logging.Logger]
EntityT = TypeVar("EntityT", bound=Entity)


class EntitiesMismatchException(Exception):
    pass


class Task(ABC):
    """
    Abstract of a Task. Provides utility functions for common calculations
    that can be helfull when creating metrics.
    Specific child classes should implement:
    - validate_config
    - calculate_score
    """

    def __init__(
        self,
        logger: loggers_type | None = None,
    ) -> None:
        if logger:
            self.logger = logger
        else:
            self.logger = logging.getLogger(__name__)

    @property
    @abstractmethod
    def system_prompt(self) -> str:
        pass

    @property
    @abstractmethod
    def task_prompt(self) -> str:
        """
        Returns the task instruction - the prompt that will be passed to agent
        """
        pass

    @abstractmethod
    def validate_config(self, simulation_config: SimulationConfig) -> bool:
        """Task should be able to verify if given config is suitable for specific task

        Args:
            simulation_config (SimulationConfig): initial scene setup
        Returns:
            bool: True is suitable, False otherwise
        """
        pass

    @abstractmethod
    def calculate_score(
        self, simulation_bridge: SimulationBridge[SimulationConfigT]
    ) -> float:
        """
        Calculate the task score based on the simulation information.

        Parameters
        ----------
        simulation_bridge : SimulationBridge[SimulationConfigT]
            The simulation bridge used to retrieve simulation data.

        Returns
        -------
        float
            A score between 0.0 and 1.0.
        """
        pass

    def filter_entities_by_object_type(
        self, entities: List[EntityT], object_types: List[str]
    ) -> List[EntityT]:
        """
        Filter and return only the entities that match the provided prefab types.

        Parameters
        ----------
        entities : List[EntityT]
            The list of entities to filter.
        object_types : List[str]
            The allowed object types.

        Returns
        -------
        List[EntityT]
            A list of entities whose prefab_name is in object_types.
        """
        return [ent for ent in entities if ent.prefab_name in object_types]

    def euclidean_distance(self, pos1: Pose, pos2: Pose) -> float:
        """Calculate euclidean distance between 2 positions"""
        return (
            (pos1.position.x - pos2.position.x) ** 2
            + (pos1.position.y - pos2.position.y) ** 2
            + (pos1.position.z - pos2.position.z) ** 2
        ) ** 0.5

    def is_adjacent(self, pos1: Pose, pos2: Pose, threshold_distance: float):
        """
        Check if two positions are adjacent, based on a threshold distance.

        Parameters
        ----------
        pos1 : Pose
            The first position.
        pos2 : Pose
            The second position.
        threshold_distance : float
            The maximum allowed distance for the positions to be considered adjacent.

        Returns
        -------
        bool
            True if the Euclidean distance between pos1 and pos2 is less than threshold_distance, False otherwise.
        """
        return self.euclidean_distance(pos1, pos2) < threshold_distance

    def is_adjacent_to_any(
        self, pos1: Pose, positions: List[Pose], threshold_distance: float
    ) -> bool:
        """
        Check if a position is adjacent to any position in a given list.

        Parameters
        ----------
        pos1 : Pose
            The position to check.
        positions : List[Pose]
            A list of positions to compare against.
        threshold_distance : float
            The distance threshold for adjacency.

        Returns
        -------
        bool
            True if pos1 is adjacent to any position in positions, False otherwise.
        """

        return any(
            self.is_adjacent(pos1, pos2, threshold_distance) for pos2 in positions
        )

    def count_adjacent(self, positions: List[Pose], threshold_distance: float) -> int:
        """
        Count how many positions in the list are adjacent to at least one other position.

        Parameters
        ----------
        positions : List[Pose]
            A list of positions.
        threshold_distance : float
            The distance threshold to determine adjacency.

        Returns
        -------
        int
            The count of positions that are adjacent to at least one other position.
        """
        adjacent_count = 0

        for i, p1 in enumerate(positions):
            for j, p2 in enumerate(positions):
                if i != j:
                    if self.is_adjacent(p1, p2, threshold_distance):
                        adjacent_count += 1
                        break

        return adjacent_count

    def build_neighbourhood_list(
        self, entities: List[EntityT], threshold_distance: float = 0.15
    ) -> Dict[EntityT, List[EntityT]]:
        """
        Build a neighbourhood list assigning a list of neighbours to every entity based on a threshold distance.

        Parameters
        ----------
        entities : List[EntityT]
            The list of entities.
        threshold_distance : float, optional
            The maximum distance between entities to consider them neighbours. Default is 0.15.

        Returns
        -------
        Dict[EntityT, List[EntityT]]
            A dictionary mapping each entity to a list of neighbouring entities.
        """
        neighbourhood_graph: Dict[EntityT, List[EntityT]] = {
            entity: [] for entity in entities
        }
        for entity in entities:
            neighbourhood_graph[entity] = [
                other
                for other in entities
                if entity != other
                and self.is_adjacent(
                    entity.pose.pose, other.pose.pose, threshold_distance
                )
            ]
        return neighbourhood_graph

    def group_entities_by_type(
        self, entities: List[EntityT]
    ) -> Dict[str, List[EntityT]]:
        """
        Group entities by their prefab type.

        Parameters
        ----------
        entities : List[EntityT]
            The list of entities to group.

        Returns
        -------
        Dict[str, List[EntityT]]
            A dictionary with keys as prefab names and values as lists of entities of that type.
        """
        entities_by_type: Dict[str, List[EntityT]] = defaultdict(list)
        for entity in entities:
            entities_by_type[entity.prefab_name].append(entity)
        return entities_by_type

    def check_neighbourhood_types(
        self,
        neighbourhood: List[EntityT],
        allowed_types: List[str],
    ) -> bool:
        """
        Check if all entities in the neighbourhood are of the allowed types.

        Parameters
        ----------
        neighbourhood : List[EntityT]
            The list of neighbouring entities.
        allowed_types : List[str]
            The allowed prefab types.

        Returns
        -------
        bool
            True if the neighbourhood is empty or if all neighbours have a prefab_name in allowed_types, False otherwise.
        """
        return not neighbourhood or all(
            adj.prefab_name in allowed_types for adj in neighbourhood
        )

    def find_clusters(
        self, neighbourhood_list: Dict[EntityT, List[EntityT]]
    ) -> List[List[EntityT]]:
        """
        Identify clusters of entities using a DFS algorithm.

        Each connected component in the neighbourhood graph is considered a cluster.
        Lone entities are counted as their own cluster.

        Parameters
        ----------
        neighbourhood_list : Dict[EntityT, List[EntityT]]
            A dictionary mapping entities to their list of neighbours.

        Returns
        -------
        List[List[EntityT]]
            A list of clusters, where each cluster is a list of connected entities.
        """
        visited: Set[EntityT] = set()
        clusters: List[List[EntityT]] = []

        def dfs(node: EntityT, cluster: List[EntityT]):
            visited.add(node)
            cluster.append(node)
            for neighbor in neighbourhood_list.get(node, []):
                if neighbor not in visited:
                    dfs(neighbor, cluster)

        for node in neighbourhood_list.keys():
            if node not in visited:
                component: List[EntityT] = []
                dfs(node, component)
                clusters.append(component)

        return clusters

    def group_entities_along_z_axis(
        # NOTE (jmatejcz) figure out how to group by other coords and orientation, without reapeting code
        self,
        entities: List[EntityT],
        margin: float,
    ) -> List[List[EntityT]]:
        """
        Group entities that are aligned along the z axis based on their x and y coordinates.

        Entities are first sorted by their x and y coordinates. Then, each entity is added to an existing group
        if its (x, y) distance from the first entity in the group is within the specified margin.
        Otherwise, a new group is created.

        Example
        ----------
        You have 2 separate vertical towers of cubes.
        In that case method will return 2 groups of entities, one for each tower.

        Parameters
        ----------
        entities : List[EntityT]
            The list of entities to group.
        margin : float
            The maximum allowable Euclidean distance in the x-y plane to consider entities as part of the same group.

        Returns
        -------
        List[List[EntityT]]
            A list of groups (clusters) of entities.
        """

        entities = sorted(
            entities,
            key=lambda ent: (ent.pose.pose.position.x, ent.pose.pose.position.y),
        )

        groups: List[List[EntityT]] = []
        for entity in entities:
            placed = False
            for group in groups:
                dx = group[0].pose.pose.position.x - entity.pose.pose.position.x
                dy = group[0].pose.pose.position.y - entity.pose.pose.position.y
                if math.sqrt(dx * dx + dy * dy) <= margin:
                    group.append(entity)
                    placed = True
                    break
            if not placed:
                groups.append([entity])
        return groups


class ManipulationTask(Task, ABC):
    """
    Common class for manipulaiton tasks
    obj_types variable represents object types that will be considered as the subject of the task.
    That means that based on positions of these objects simulation config will be evaluated
    and score will be calculated.

    Example
    -------
        MoveObjectsToLeftTask with 'carrot' as objects type, will check if carrtos are present
        and then calculated score based on how many carrots were moved to the left side
    """

    obj_types: List[str] = []

    @property
    def system_prompt(self) -> str:
        return """
    You are a robotic arm with interfaces to detect and manipulate objects.
    Here are the coordinates information:
    x - front to back (positive is forward)
    y - left to right (positive is right)
    z - up to down (positive is up)
    Before starting the task, make sure to grab the camera image to understand the environment.
    """

    @abstractmethod
    def check_if_required_objects_present(
        self, simulation_config: SimulationConfig
    ) -> bool:
        """
        Check if the required objects are present in the simulation configuration.

        Returns
        -------
        bool
            True if the required objects are present, False otherwise.
        """
        return True

    def check_if_any_placed_incorrectly(
        self, simulation_config: SimulationConfig
    ) -> bool:
        """
        Check if any object is placed incorrectly in the simulation configuration.
        Save number of initially correctly and incorrectly placed objects for
        future calculations

        Returns
        -------
        bool
            True if at least one object is placed incorrectly, False otherwise.
        """
        _, incorrect = self.calculate_correct(entities=simulation_config.entities)
        return incorrect > 0

    def validate_config(self, simulation_config: SimulationConfig) -> bool:
        """
        Validate the simulation configuration.

        Checks whether the required objects are present and if any of them is placed incorrectly.
        If these conditions are not met, the task should not be run with this configuration.

        Parameters
        ----------
        simulation_config : SimulationConfig
            The simulation configuration to validate.

        Returns
        -------
        bool
            True if the configuration is valid, False otherwise.
        """

        if self.check_if_required_objects_present(
            simulation_config=simulation_config
        ) and self.check_if_any_placed_incorrectly(simulation_config=simulation_config):
            return True
        else:
            return False

    @abstractmethod
    def calculate_correct(
        self, entities: List[Entity] | List[SpawnedEntity]
    ) -> Tuple[int, int]:
        """Method to calculate how many objects are placed correctly

        Parameters
        ----------
        entities : List[Entity]
            list of ALL entities present in the simulaiton scene

        Returns
        -------
        Tuple[int, int]
            first int HAVE TO be number of correctly placed objects, second int - number of incorrectly placed objects
        """
        pass

    def calculate_current_placements(
        self, simulation_bridge: SimulationBridge[SimulationConfigT]
    ) -> tuple[int, int]:
        """
        Get the current placements of objects in the simulation
        and calculated their current placements

        Parameters
        ----------
        simulation_bridge : SimulationBridge[SimulationConfigT]
            The simulation bridge containing the current scene state.

        Returns
        -------
        tuple[int, int]
            A tuple where the first element is the number of currently correctly placed objects
            and the second element is the number of currently incorrectly placed objects.
        """
        scene_state = simulation_bridge.get_scene_state()
        current_correct, current_incorrect = self.calculate_correct(
            entities=scene_state.entities
        )

        self.logger.info(  # type: ignore
            f"Currently correctly placed objects: {current_correct}, Currenlty incorrectly placed objects: {current_incorrect}"
        )
        return current_correct, current_incorrect

    def calculate_score(
        self, simulation_bridge: SimulationBridge[SimulationConfig]
    ) -> float:
        """
        Calculate the task score based on the difference between initial and current placements.

        The score ranges from 0.0 to 1.0, where 0.0 indicates that the initial placements
        remain unchanged (or got worse), and 1.0 indicates perfect placements relative to the initial ones.
        The score is computed as the improvement in the number of correctly placed objects
        divided by the number of initially incorrectly placed objects.

        Parameters
        ----------
        simulation_bridge : SimulationBridge[SimulationConfig]
            The simulation bridge that provides access to the current scene state.

        Returns
        -------
        float
            The calculated score, ranging from 0.0 to 1.0.

        Raises
        ------
        EntitiesMismatchException
            If the total number of initial entities does not match the total number of current entities.
        """
        initially_correct, initially_incorrect = self.calculate_correct(
            entities=simulation_bridge.spawned_entities
        )
        self.logger.info(  # type: ignore
            f"Objects placed correctly in simulation config: {initially_correct}, incorrectly: {initially_incorrect}"
        )
        current_correct, current_incorrect = self.calculate_current_placements(
            simulation_bridge
        )

        initial_objects_num = initially_correct + initially_incorrect
        current_objects_num = current_correct + current_incorrect
        if initial_objects_num == 0:
            return 1.0
        elif initial_objects_num != current_objects_num:
            raise EntitiesMismatchException(
                f"number of initial entities does not match current entities number, initially: {initially_correct + initially_incorrect}, current: {current_correct + current_incorrect}"
            )
        else:
            corrected = current_correct - initially_correct
            score = max(0.0, corrected / initially_incorrect)

            self.logger.info(f"Calculated score: {score:.2f}")  # type: ignore
            return score



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from rai_bench.manipulation_o3de.tasks.build_tower_task import (
    BuildCubeTowerTask,
)
from rai_bench.manipulation_o3de.tasks.group_objects_task import (
    GroupObjectsTask,
)
from rai_bench.manipulation_o3de.tasks.move_object_to_left_task import (
    MoveObjectsToLeftTask,
)
from rai_bench.manipulation_o3de.tasks.place_at_coord_task import (
    PlaceObjectAtCoordTask,
)
from rai_bench.manipulation_o3de.tasks.place_cubes_task import (
    PlaceCubesTask,
)
from rai_bench.manipulation_o3de.tasks.rotate_object_task import (
    RotateObjectTask,
)

__all__ = [
    "BuildCubeTowerTask",
    "GroupObjectsTask",
    "MoveObjectsToLeftTask",
    "PlaceCubesTask",
    "PlaceObjectAtCoordTask",
    "RotateObjectTask",
]



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from typing import List, Sequence

from rai.tools.ros2 import MoveToPointToolInput

from rai_bench.tool_calling_agent.interfaces import (
    Task,
)
from rai_bench.tool_calling_agent.tasks.basic import (
    GetAllROS2DepthCamerasTask,
    GetAllROS2RGBCamerasTask,
    GetROS2DepthCameraTask,
    GetROS2RGBCameraTask,
    GetROS2TopicsTask,
)
from rai_bench.tool_calling_agent.tasks.custom_interfaces import (
    PublishROS2HRIMessageTextTask,
)
from rai_bench.tool_calling_agent.tasks.manipulation import (
    MoveToPointTask,
)
from rai_bench.tool_calling_agent.tasks.navigation import (
    MoveToBedTask,
    MoveToFrontTask,
    NavigateToPointTask,
    SpinAroundTask,
)
from rai_bench.tool_calling_agent.tasks.spatial import (
    BoolImageTask,
    BoolImageTaskInput,
)
from rai_bench.tool_calling_agent.tasks.subtasks import (
    CheckActionFieldsToolCallSubTask,
    CheckArgsToolCallSubTask,
    CheckTopicFieldsToolCallSubTask,
)
from rai_bench.tool_calling_agent.validators import (
    NotOrderedCallsValidator,
    OrderedCallsValidator,
)

true_response_inputs: List[BoolImageTaskInput] = [
    BoolImageTaskInput(
        question="Is the door on the left from the desk?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_1.jpg"],
    ),
    BoolImageTaskInput(
        question="Is the light on in the room?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_2.jpg"],
    ),
    BoolImageTaskInput(
        question="Do you see the plant?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_2.jpg"],
    ),
    BoolImageTaskInput(
        question="Are there any pictures on the wall?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_3.jpg"],
    ),
    BoolImageTaskInput(
        question="Are there 3 pictures on the wall?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_4.jpg"],
    ),
    BoolImageTaskInput(
        question="Is there a plant behind the rack?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_5.jpg"],
    ),
    BoolImageTaskInput(
        question="Is there a pillow on the armchain?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_7.jpg"],
    ),
]
false_response_inputs: List[BoolImageTaskInput] = [
    BoolImageTaskInput(
        question="Is the door open?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_1.jpg"],
    ),
    BoolImageTaskInput(
        question="Is someone in the room?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_1.jpg"],
    ),
    BoolImageTaskInput(
        question="Do you see the plant?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_3.jpg"],
    ),
    BoolImageTaskInput(
        question="Are there 4 pictures on the wall?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_4.jpg"],
    ),
    BoolImageTaskInput(
        question="Is there a rack on the left from the sofa?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_4.jpg"],
    ),
    BoolImageTaskInput(
        question="Is there a plant on the right from the window?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_6.jpg"],
    ),
    BoolImageTaskInput(
        question="Is there a red pillow on the armchair?",
        images_paths=["src/rai_bench/rai_bench/examples/images/image_7.jpg"],
    ),
]

########## SUBTASKS #######################################################################################
get_topics_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_topics_names_and_types", expected_args={}
)

color_image_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_image", expected_args={"topic": "/camera_image_color"}
)
depth_image_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_image", expected_args={"topic": "/camera_image_depth"}
)
color_image5_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_image", expected_args={"topic": "/color_image5"}
)
depth_image5_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_image", expected_args={"topic": "/depth_image5"}
)

receive_robot_desc_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="receive_ros2_message",
    expected_args={"topic": "/robot_description"},
)

move_to_point_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="move_to_point",
    expected_args={"x": 1.0, "y": 2.0, "z": 3.0, "task": "grab"},
)

pub_HRIMessage_text_subtask = CheckTopicFieldsToolCallSubTask(
    expected_tool_name="publish_ros2_message",
    expected_topic="/to_human",
    expected_message_type="rai_interfaces/msg/HRIMessage",
    expected_fields={"text": "Hello!"},
)

get_tohuman_interface_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="get_ros2_message_interface",
    expected_args={"msg_type": "rai_interfaces/msg/HRIMessage"},
)


return_true_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="return_bool_response", expected_args={"response": True}
)
return_false_subtask = CheckArgsToolCallSubTask(
    expected_tool_name="return_bool_response", expected_args={"response": False}
)

start_nav_action_subtask = CheckActionFieldsToolCallSubTask(
    expected_tool_name="start_ros2_action",
    expected_action="/navigate_to_pose",
    expected_action_type="nav2_msgs/action/NavigateToPose",
    expected_fields={
        "pose": {
            "header": {"frame_id": "map"},
            "pose": {
                "position": {"x": 2.0, "y": 2.0, "z": 0.0},
            },
        },
    },
)
start_spin_action_subtask = CheckActionFieldsToolCallSubTask(
    expected_tool_name="start_ros2_action",
    expected_action="/spin",
    expected_action_type="nav2_msgs/action/Spin",
    expected_fields={"target_yaw": 3},
)
start_move_front_action_subtask = CheckActionFieldsToolCallSubTask(
    expected_tool_name="start_ros2_action",
    expected_action="/drive_on_heading",
    expected_action_type="nav2_msgs/action/DriveOnHeading",
    expected_fields={
        "target": {"y": 0.0, "z": 0.0},
    },
)
######### VALIDATORS #########################################################################################
topics_ord_val = OrderedCallsValidator(subtasks=[get_topics_subtask])
topics_and_color_image_ord_val = OrderedCallsValidator(
    subtasks=[
        get_topics_subtask,
        color_image_subtask,
    ]
)
color_image_ord_val = OrderedCallsValidator(subtasks=[color_image_subtask])
depth_image_ord_val = OrderedCallsValidator(subtasks=[depth_image_subtask])
all_color_images_notord_val = NotOrderedCallsValidator(
    subtasks=[color_image_subtask, color_image5_subtask]
)
all_depth_images_notord_val = NotOrderedCallsValidator(
    subtasks=[depth_image_subtask, depth_image5_subtask]
)

move_to_point_ord_val = OrderedCallsValidator(subtasks=[move_to_point_subtask])

pub_HRIMessage_text_ord_val = OrderedCallsValidator(
    subtasks=[pub_HRIMessage_text_subtask]
)

list_topic_get_interface_publish_ord_val = OrderedCallsValidator(
    subtasks=[
        get_topics_subtask,
        get_tohuman_interface_subtask,
        pub_HRIMessage_text_subtask,
    ]
)

ret_true_ord_val = OrderedCallsValidator(subtasks=[return_true_subtask])
ret_false_ord_val = OrderedCallsValidator(subtasks=[return_false_subtask])

start_navigate_action_ord_val = OrderedCallsValidator(
    subtasks=[start_nav_action_subtask]
)
start_spin_action_ord_val = OrderedCallsValidator(subtasks=[start_spin_action_subtask])
move_ahead_ord_val = OrderedCallsValidator(subtasks=[start_move_front_action_subtask])

######### TASKS ############################################################################################
basic_tasks: List[Task] = [
    # 3 options to validate same task:
    # most strict, agent has to call both tool correctly to pass this validator
    GetROS2RGBCameraTask(validators=[topics_and_color_image_ord_val]),
    # verifing only if the GetCameraImage call was made properly
    GetROS2RGBCameraTask(validators=[color_image_ord_val]),
    # Soft verification. verifing in separate vaidators the list topic and get image.
    #  agent can get 0.5 score by only calling list topics
    GetROS2RGBCameraTask(validators=[topics_ord_val, color_image_ord_val]),
    # we can also add extra tool calls to allow model to correct itself
    GetROS2RGBCameraTask(
        validators=[topics_ord_val, color_image_ord_val], extra_tool_calls=3
    ),
    GetROS2TopicsTask(validators=[topics_ord_val]),
    GetROS2DepthCameraTask(validators=[depth_image_ord_val]),
    GetAllROS2RGBCamerasTask(validators=[all_color_images_notord_val]),
    GetAllROS2DepthCamerasTask(validators=[all_depth_images_notord_val]),
]
manipulation_tasks: List[Task] = [
    MoveToPointTask(
        move_to_tool_input=MoveToPointToolInput(x=1.0, y=2.0, z=3.0, task="grab"),
        validators=[move_to_point_ord_val],
    ),
    MoveToPointTask(
        move_to_tool_input=MoveToPointToolInput(x=1.2, y=2.3, z=3.4, task="drop"),
        validators=[move_to_point_ord_val],
    ),
]

custom_interfaces_tasks: List[Task] = [
    PublishROS2HRIMessageTextTask(
        topic="/to_human",
        text="Hello!",
        validators=[pub_HRIMessage_text_ord_val],
        extra_tool_calls=2,
    ),
    PublishROS2HRIMessageTextTask(
        topic="/to_human",
        text="Hello!",
        validators=[list_topic_get_interface_publish_ord_val],
    ),
]

true_spatial_tasks: List[Task] = [
    BoolImageTask(
        task_input=input_item, validators=[ret_true_ord_val], extra_tool_calls=0
    )
    for input_item in true_response_inputs
]
false_spatial_tasks: List[Task] = [
    BoolImageTask(task_input=input_item, validators=[ret_false_ord_val])
    for input_item in false_response_inputs
]

navigation_tasks: List[Task] = [
    NavigateToPointTask(validators=[start_navigate_action_ord_val], extra_tool_calls=5),
    SpinAroundTask(validators=[start_spin_action_ord_val], extra_tool_calls=5),
    MoveToBedTask(validators=[move_ahead_ord_val], extra_tool_calls=5),
    MoveToFrontTask(validators=[move_ahead_ord_val], extra_tool_calls=5),
]


def get_basic_tasks(extra_tool_calls: int = 0) -> List[Task]:
    return [
        # 3 options to validate same task:
        # most strict, agent has to call both tool correctly to pass this validator
        GetROS2RGBCameraTask(
            validators=[topics_and_color_image_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        # verifing only if the GetCameraImage call was made properly
        GetROS2RGBCameraTask(
            validators=[color_image_ord_val], extra_tool_calls=extra_tool_calls
        ),
        # Soft verification. verifing in separate vaidators the list topic and get image.
        #  agent can get 0.5 score by only calling list topics
        GetROS2RGBCameraTask(
            validators=[topics_ord_val, color_image_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        # we can also add extra tool calls to allow model to correct itself
        GetROS2RGBCameraTask(
            validators=[topics_ord_val, color_image_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        GetROS2TopicsTask(
            validators=[topics_ord_val], extra_tool_calls=extra_tool_calls
        ),
        GetROS2DepthCameraTask(
            validators=[depth_image_ord_val], extra_tool_calls=extra_tool_calls
        ),
        GetAllROS2RGBCamerasTask(
            validators=[all_color_images_notord_val], extra_tool_calls=extra_tool_calls
        ),
        GetAllROS2DepthCamerasTask(
            validators=[all_depth_images_notord_val], extra_tool_calls=extra_tool_calls
        ),
    ]


def get_navigation_tasks(extra_tool_calls: int = 0) -> List[Task]:
    return [
        NavigateToPointTask(
            validators=[start_navigate_action_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        SpinAroundTask(
            validators=[start_spin_action_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        MoveToBedTask(
            validators=[move_ahead_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        MoveToFrontTask(
            validators=[move_ahead_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
    ]


def get_manipulation_tasks(extra_tool_calls: int = 0) -> List[Task]:
    return [
        MoveToPointTask(
            move_to_tool_input=MoveToPointToolInput(x=1.0, y=2.0, z=3.0, task="grab"),
            validators=[move_to_point_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        MoveToPointTask(
            move_to_tool_input=MoveToPointToolInput(x=1.2, y=2.3, z=3.4, task="drop"),
            validators=[move_to_point_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
    ]


def get_custom_interfaces_tasks(extra_tool_calls: int = 0) -> List[Task]:
    return [
        PublishROS2HRIMessageTextTask(
            topic="/to_human",
            text="Hello!",
            validators=[pub_HRIMessage_text_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
        PublishROS2HRIMessageTextTask(
            topic="/to_human",
            text="Hello!",
            validators=[list_topic_get_interface_publish_ord_val],
            extra_tool_calls=extra_tool_calls,
        ),
    ]


def get_spatial_tasks(extra_tool_calls: int = 0) -> Sequence[Task]:
    true_tasks = [
        BoolImageTask(
            task_input=input_item,
            validators=[ret_true_ord_val],
            extra_tool_calls=extra_tool_calls,
        )
        for input_item in true_response_inputs
    ]
    false_tasks = [
        BoolImageTask(
            task_input=input_item,
            validators=[ret_false_ord_val],
            extra_tool_calls=extra_tool_calls,
        )
        for input_item in false_response_inputs
    ]
    return true_tasks + false_tasks


def get_all_tasks(extra_tool_calls: int = 0) -> List[Task]:
    tasks: List[Task] = []
    tasks += get_basic_tasks(extra_tool_calls=extra_tool_calls)
    tasks += get_custom_interfaces_tasks(extra_tool_calls=extra_tool_calls)
    tasks += get_manipulation_tasks(extra_tool_calls=extra_tool_calls)
    tasks += get_navigation_tasks(extra_tool_calls=extra_tool_calls)
    tasks += get_spatial_tasks(extra_tool_calls=extra_tool_calls)

    random.shuffle(tasks)
    return tasks



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from abc import ABC, abstractmethod
from typing import Dict, List

import inflect
from langchain_core.tools import BaseTool
from rai.tools.ros2 import MoveToPointToolInput
from rai.types import Point

from rai_bench.tool_calling_agent.interfaces import Task, Validator
from rai_bench.tool_calling_agent.mocked_tools import (
    MockGetObjectPositionsTool,
    MockGetROS2TopicsNamesAndTypesTool,
    MockMoveToPointTool,
)

loggers_type = logging.Logger

PROACTIVE_ROS2_EXPERT_SYSTEM_PROMPT = """
        You are a robotic arm with interfaces to detect and manipulate objects.
        Here are the coordinates information:
        x - front to back (positive is forward)
        y - left to right (positive is right)
        z - up to down (positive is up).
        """


class TaskParametrizationError(Exception):
    """Exception raised when the task parameters are not valid."""

    pass


class ManipulationTask(Task, ABC):
    @property
    def type(self) -> str:
        return "manipulation"

    def get_system_prompt(self) -> str:
        return PROACTIVE_ROS2_EXPERT_SYSTEM_PROMPT


class GrabTask(ManipulationTask, ABC):
    def __init__(
        self,
        objects: Dict[str, List[Point]],
        object_to_grab: str,
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators,
            extra_tool_calls=extra_tool_calls,
            logger=logger,
        )
        self.objects = objects
        self.object_to_grab = object_to_grab
        self._verify_args()

    @abstractmethod
    def _verify_args(self) -> None:
        pass

    @property
    def available_tools(self) -> List[BaseTool]:
        return [
            MockGetROS2TopicsNamesAndTypesTool(
                mock_topics_names_and_types=[
                    "topic: /attached_collision_object\ntype: moveit_msgs/msg/AttachedCollisionObject\n",
                    "topic: /camera_image_color\ntype: sensor_msgs/msg/Image\n",
                    "topic: /camera_image_depth\ntype: sensor_msgs/msg/Image\n",
                    "topic: /clock\ntype: rosgraph_msgs/msg/Clock\n",
                    "topic: /collision_object\ntype: moveit_msgs/msg/CollisionObject\n",
                    "topic: /color_camera_info\ntype: sensor_msgs/msg/CameraInfo\n",
                ]
            ),
            MockGetObjectPositionsTool(
                target_frame="panda_link0",
                source_frame="RGBDCamera5",
                camera_topic="/color_image5",
                depth_topic="/depth_image5",
                camera_info_topic="/color_camera_info5",
                mock_objects=self.objects,
            ),
            MockMoveToPointTool(manipulator_frame="panda_link0"),
        ]


class MoveToPointTask(ManipulationTask):
    complexity = "easy"

    def __init__(
        self,
        move_to_tool_input: MoveToPointToolInput,
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators, extra_tool_calls=extra_tool_calls, logger=logger
        )

        self.move_to_tool_input = move_to_tool_input

    @property
    def available_tools(self) -> List[BaseTool]:
        return [
            MockGetROS2TopicsNamesAndTypesTool(
                mock_topics_names_and_types=[
                    "topic: /pointcloud\ntype: sensor_msgs/msg/PointCloud2\n",
                    "topic: /robot_description\ntype: std_msgs/msg/String\n",
                    "topic: /rosout\ntype: rcl_interfaces/msg/Log\n",
                    "topic: /tf\ntype: tf2_msgs/msg/TFMessage\n",
                ]
            ),
            MockMoveToPointTool(manipulator_frame="base_link"),
        ]

    def get_prompt(self) -> str:
        return f"Move the arm to a point x={self.move_to_tool_input.x}, y={self.move_to_tool_input.y}, z={self.move_to_tool_input.z} to {self.move_to_tool_input.task} an object."


class GetObjectPositionsTask(ManipulationTask):
    complexity = "easy"

    def __init__(
        self,
        objects: Dict[str, List[Point]],
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators, extra_tool_calls=extra_tool_calls, logger=logger
        )
        """Task to get the positions of the objects

        Examples
        --------
        objects = {
            "banana": [(0.1, 0.2, 0.3), (0.4, 0.5, 0.6)],
            "cube": [(0.7, 0.8, 0.9)],
        }
        """
        self.objects = objects

    @property
    def available_tools(self) -> List[BaseTool]:
        return [
            MockGetROS2TopicsNamesAndTypesTool(
                mock_topics_names_and_types=[
                    "topic: /pointcloud\ntype: sensor_msgs/msg/PointCloud2\n",
                    "topic: /robot_description\ntype: std_msgs/msg/String\n",
                    "topic: /rosout\ntype: rcl_interfaces/msg/Log\n",
                    "topic: /tf\ntype: tf2_msgs/msg/TFMessage\n",
                ]
            ),
            MockGetObjectPositionsTool(mock_objects=self.objects),
        ]

    def get_prompt(self) -> str:
        """Generates a prompt based on the objects provided in the task. If there is more than one object, the object in the prompt will be pluralized.
        Returns:
            str: Formatted prompt for the task
        """
        inflector = inflect.engine()
        object_counts = {obj: len(positions) for obj, positions in self.objects.items()}
        formatted_objects = [
            inflector.plural(obj) if count > 1 else obj
            for obj, count in object_counts.items()
        ]
        if len(formatted_objects) > 1:
            objects_list = (
                ", ".join(formatted_objects[:-1]) + f", and {formatted_objects[-1]}"
            )
        else:
            objects_list = formatted_objects[0]
        return f"Get the {objects_list} positions."


class GrabExistingObjectTask(GrabTask):
    complexity = "medium"
    """
    Task to grab an object.

    Parameters
    ----------
    objects : Dict[str, List[dict[str, float]]]
        Dictionary of object types and their positions.
    object_to_grab : str
        The object to be grabbed (must have a single position).
    """

    def get_prompt(self) -> str:
        return f"Grab {self.object_to_grab}."

    def _verify_args(self):
        if self.object_to_grab not in self.objects:
            error_message = f"Requested object to grab {self.object_to_grab} is not present in defined objects: {self.objects}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)

        if len(self.objects[self.object_to_grab]) > 1:
            error_message = f"Requested object to grab {self.object_to_grab} has more than one position in defined objects: {self.objects[self.object_to_grab]}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)


class GrabNotExistingObjectTask(GrabTask):
    complexity = "medium"
    """
    Task to attempt grabbing an object that does not exist.

    Parameters
    ----------
    objects : Dict[str, List[dict[str, float]]]
        Available objects and their positions.
    object_to_grab : str
        Object that should not be present in the list.
    """

    def get_prompt(self) -> str:
        return f"Grab {self.object_to_grab}."

    def _verify_args(self):
        if self.object_to_grab in self.objects:
            error_message = f"Requested object to grab {self.object_to_grab} is present in defined objects: {self.objects} but should not be."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)


class MoveExistingObjectLeftTask(GrabTask):
    """Task to move an existing object to the left.

    Parameters
    ----------
    objects : Dict[str, List[dict[str, float]]]
        Dictionary containing the object types and their positions. Object type should be passed as singular.
    object_to_grab : str
        Object type should be passed as singular. Object to be grabbed should be defined in the objects argument with only one instance (one position).
    logger : loggers_type | None, optional
        Logger, by default None

    Examples
    --------
    objects = {
        "banana": [(0.1, 0.2, 0.3), (0.4, 0.5, 0.6)],
        "cube": [(0.7, 0.8, 0.9)],
    }
    object_to_grab = "cube"
    """

    complexity = "medium"

    def get_prompt(self) -> str:
        return f"Move {self.object_to_grab} 20 cm to the left."

    def _verify_args(self):
        if self.object_to_grab not in self.objects:
            error_message = f"Requested object to grab {self.object_to_grab} is not present in defined objects: {self.objects}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)

        if len(self.objects[self.object_to_grab]) > 1:
            error_message = f"Requested object to grab {self.object_to_grab} has more than one position in defined objects: {self.objects[self.object_to_grab]}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)


class MoveExistingObjectFrontTask(GrabTask):
    """Task to move an existing object to the front

    Parameters
    ----------
    objects : Dict[str, List[dict[str, float]]]
        Dictionary containing the object types and their positions. Object type should be passed as singular.
    object_to_grab : str
        Object to grab. Object type should be passed as singular. Object to be grabbed should be defined in the objects argument with only one instance (one position).
    logger : loggers_type | None, optional
        Logger, by default None
    """

    complexity = "medium"

    def get_prompt(self) -> str:
        return f"Move {self.object_to_grab} 60 cm to the front."

    def _verify_args(self):
        if self.object_to_grab not in self.objects:
            error_message = f"Requested object to grab {self.object_to_grab} is not present in defined objects: {self.objects}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)

        if len(self.objects[self.object_to_grab]) > 1:
            error_message = f"Requested object to grab {self.object_to_grab} has more than one position in defined objects: {self.objects[self.object_to_grab]}."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)


class SwapObjectsTask(Task):
    """Task to swap objects

    Parameters
    ----------
    objects : Dict[str, List[Dict[str, float]]]
        Dictionary containing the object types and their positions. Object type should be passed as singular.
    objects_to_swap : List[str]
        Objects to be swapped. Object type should be passed as singular. Objects to be swapped should be defined in the objects argument with only one instance (one position).
    logger : loggers_type | None, optional
        Logger, by default None

    Examples
    --------
    objects = {
        "banana": [(0.1, 0.2, 0.1)],
        "cube": [(0.7, 0.8, 0.1)],
        "apple": [(0.3, 0.4, 0.1), (0.5, 0.6, 0.1)],

    }
    objects_to_swap = ["cube", "banana"]
    """

    complexity = "hard"

    def __init__(
        self,
        objects: Dict[str, List[Point]],
        objects_to_swap: str,
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators,
            extra_tool_calls=extra_tool_calls,
            logger=logger,
        )
        self.objects = objects
        self.objects_to_swap = objects_to_swap
        self._verify_args()

    @property
    def available_tools(self) -> List[BaseTool]:
        return [
            MockGetROS2TopicsNamesAndTypesTool(
                mock_topics_names_and_types=[
                    "topic: /attached_collision_object\ntype: moveit_msgs/msg/AttachedCollisionObject\n",
                    "topic: /camera_image_color\ntype: sensor_msgs/msg/Image\n",
                    "topic: /camera_image_depth\ntype: sensor_msgs/msg/Image\n",
                    "topic: /clock\ntype: rosgraph_msgs/msg/Clock\n",
                    "topic: /collision_object\ntype: moveit_msgs/msg/CollisionObject\n",
                    "topic: /color_camera_info\ntype: sensor_msgs/msg/CameraInfo\n",
                ]
            ),
            MockGetObjectPositionsTool(
                target_frame="panda_link0",
                source_frame="RGBDCamera5",
                camera_topic="/color_image5",
                depth_topic="/depth_image5",
                camera_info_topic="/color_camera_info5",
                mock_objects=self.objects,
            ),
            MockMoveToPointTool(manipulator_frame="panda_link0"),
        ]

    def _verify_args(self):
        for obj in self.objects_to_swap:
            if obj not in self.objects:
                error_message = f"Requested object to swap {obj} is not present in defined objects: {self.objects}."
                self.logger.error(msg=error_message)
                raise TaskParametrizationError(error_message)
            if len(self.objects[obj]) != 1:
                error_message = f"Number of positions for object to swap ({obj}) should be equal to 1."
                self.logger.error(msg=error_message)
                raise TaskParametrizationError(error_message)
        if len(self.objects_to_swap) != 2:
            error_message = f"Number of requested objects to swap {len(self.objects_to_swap)} should be equal to 2."
            self.logger.error(msg=error_message)
            raise TaskParametrizationError(error_message)

    def get_prompt(self) -> str:
        return f"Move {self.objects_to_swap[0]} to the initial position of {self.objects_to_swap[1]}, and move {self.objects_to_swap[1]} to the initial position of {self.objects_to_swap[0]}."



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
from abc import abstractmethod
from typing import List

from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from rai.messages import preprocess_image

from rai_bench.tool_calling_agent.interfaces import Task, Validator

loggers_type = logging.Logger

SPATIAL_REASONING_SYSTEM_PROMPT = "You are a helpful and knowledgeable AI assistant that specializes in interpreting and analyzing visual content. Your task is to answer questions based on the images provided to you. Please response with the use of the provided tools."


class TaskParametrizationError(Exception):
    """Exception raised when the task parameters are not valid."""

    pass


class SpatialReasoningAgentTask(Task):
    """Abstract class for spatial reasoning tasks for tool calling agent."""

    def __init__(
        self,
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators,
            extra_tool_calls=extra_tool_calls,
            logger=logger,
        )
        self.expected_tools: List[BaseTool]
        self.question: str
        self.images_paths: List[str]

    @property
    def type(self) -> str:
        return "spatial_reasoning"

    @abstractmethod
    def get_images(self) -> List[str]:
        """Get the images related to the task.

        Returns
        -------
        List[str]
            List of image paths
        """
        pass

    def get_system_prompt(self) -> str:
        return SPATIAL_REASONING_SYSTEM_PROMPT


class ReturnBoolResponseToolInput(BaseModel):
    response: bool = Field(..., description="The response to the question.")


class ReturnBoolResponseTool(BaseTool):
    """Tool that returns a boolean response."""

    name: str = "return_bool_response"
    description: str = "Return a bool response to the question."
    args_schema = ReturnBoolResponseToolInput

    def _run(self, response: bool) -> bool:
        if type(response) is bool:
            return response
        raise ValueError("Invalid response type. Response must be a boolean.")


class BoolImageTaskInput(BaseModel):
    question: str = Field(..., description="The question to be answered.")
    images_paths: List[str] = Field(
        ...,
        description="List of image file paths to be used for answering the question.",
    )


class BoolImageTask(SpatialReasoningAgentTask):
    complexity = "easy"

    def __init__(
        self,
        task_input: BoolImageTaskInput,
        validators: List[Validator],
        extra_tool_calls: int = 0,
        logger: loggers_type | None = None,
    ) -> None:
        super().__init__(
            validators=validators,
            extra_tool_calls=extra_tool_calls,
            logger=logger,
        )
        self.question = task_input.question
        self.images_paths = task_input.images_paths

    @property
    def available_tools(self) -> List[BaseTool]:
        return [ReturnBoolResponseTool()]

    def get_prompt(self):
        return self.question

    def get_images(self):
        images = [preprocess_image(image_path) for image_path in self.images_paths]
        return images



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from threading import Lock
from typing import Any, Dict, List, Tuple, Type
from unittest.mock import MagicMock

import numpy as np
import numpy.typing as npt
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ValidationError
from rai.communication.ros2.connectors import ROS2Connector
from rai.communication.ros2.messages import ROS2Message
from rai.messages import MultimodalArtifact, preprocess_image
from rai.tools.ros2 import (
    CallROS2ServiceTool,
    CancelROS2ActionTool,
    GetObjectPositionsTool,
    GetROS2ActionFeedbackTool,
    GetROS2ActionIDsTool,
    GetROS2ActionResultTool,
    GetROS2ActionsNamesAndTypesTool,
    GetROS2ImageTool,
    GetROS2MessageInterfaceTool,
    GetROS2ServicesNamesAndTypesTool,
    GetROS2TopicsNamesAndTypesTool,
    MoveToPointTool,
    PublishROS2MessageTool,
    ReceiveROS2MessageTool,
    ROS2ActionToolkit,
    StartROS2ActionTool,
)
from rai_open_set_vision.tools import (
    DistanceMeasurement,
    GetDistanceToObjectsTool,
    GetGrabbingPointTool,
)


class MockGetROS2TopicsNamesAndTypesTool(GetROS2TopicsNamesAndTypesTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    mock_topics_names_and_types: list[str]

    def _run(self) -> str:
        """Mocked method that returns the mock topics and types instead of fetching from ROS2.

        Returns
        -------
        str
            Mocked output of 'get_ros2_topics_names_and_types' tool.
        """
        return "\n".join(self.mock_topics_names_and_types)


class MockGetROS2ImageTool(GetROS2ImageTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_topics: List[str]

    def _run(
        self, topic: str, timeout_sec: float = 1.0
    ) -> Tuple[str, MultimodalArtifact]:
        """Method that returns a mock black image if the passed topic is correct.

        Parameters
        ----------
        topic : str
            Topic with the image
        timeout_sec : float, optional
            Timeout in seconds, by default 1.0

        Returns
        -------
        Tuple[str, MultimodalArtifact]
            Message from the tool and the image.

        Raises
        ------
        ValueError
            If the passed topic is not correct.
        """
        if topic not in self.available_topics:
            raise ValueError(
                f"Topic {topic} is not available within {timeout_sec} seconds. Check if the topic exists."
            )
        image = self.generate_mock_image()
        return "Image received successfully", MultimodalArtifact(
            images=[preprocess_image(image)]
        )  # type: ignore

    @staticmethod
    def generate_mock_image() -> npt.NDArray[np.uint8]:
        """Generates a blank black image (480x640, RGB).

        Returns
        -------
        npt.NDArray[np.uint8]
            Blank black image
        """
        height, width = 480, 640
        blank_image = np.zeros((height, width, 3), dtype=np.uint8)
        return blank_image


class MockReceiveROS2MessageTool(ReceiveROS2MessageTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_topics: List[str]

    def _run(self, topic: str, timeout_sec: float = 1.0) -> str:
        """Method that returns a mock message if the passed topic is correct.

        Parameters
        ----------
        topic : str
            Topic to receive the message from

        Returns
        -------
        str
            Message from the tool

        Raises
        ------
        ValueError
            If the passed topic is not correct.
        """
        if topic not in self.available_topics:
            raise ValueError(
                f"Topic {topic} is not available within 1.0 seconds. Check if the topic exists."
            )
        message: ROS2Message = MagicMock(spec=ROS2Message)
        message.payload = {"mock": "payload"}
        message.metadata = {"mock": "metadata"}
        return str({"payload": message.payload, "metadata": message.metadata})


class MockMoveToPointTool(MoveToPointTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)

    def _run(self, x: float, y: float, z: float, task: str) -> str:
        """Method that return a mock message with the end effector position.

        Parameters
        ----------
        x : float
        y : float
        z : float
        task : str
            Task to perform.

        Returns
        -------
        str
            Message from the tool
        """
        return f"End effector successfully positioned at coordinates ({x:.2f}, {y:.2f}, {z:.2f}). Note: The status of object interaction (grab/drop) is not confirmed by this movement."


class MockGetObjectPositionsTool(GetObjectPositionsTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)

    # Create mock instances for the arguments
    target_frame: str = MagicMock(spec=str)
    source_frame: str = MagicMock(spec=str)
    camera_topic: str = MagicMock(spec=str)
    depth_topic: str = MagicMock(spec=str)
    camera_info_topic: str = MagicMock(spec=str)
    get_grabbing_point_tool: GetGrabbingPointTool = MagicMock(spec=GetGrabbingPointTool)
    mock_objects: dict[str, List[dict[str, float]]]

    def _run(self, object_name: str) -> str:
        """Method that returns a mock message with the object positions if the object_name is present in the mock_objects dictionary.

        Parameters
        ----------
        object_name : str
            Name of the object to get the positions of

        Returns
        -------
        str
            Message from the tool
        """
        expected_positions = self.mock_objects.get(object_name, [])
        print(f"Expected positions: {expected_positions}")
        if len([expected_positions]) == 0:
            return f"No {object_name}s detected."
        else:
            return f"Centroids of detected {object_name}s in manipulator frame: {expected_positions} Sizes of the detected objects are unknown."


class MockPublishROS2MessageTool(PublishROS2MessageTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_topics: List[str]
    available_message_types: List[str]
    available_topic_models: Dict[str, Type[BaseModel]]

    def _run(self, topic: str, message: Dict[str, Any], message_type: str) -> str:
        """
        Mocked method that simulates publihing to a topic and return a status string.

        Parameters
        ----------
        topic : str
            The name of the topic to which the message is published.
        message : Dict[str, Any]
            The content of the message as a dictionary.
        message_type : str
            The type of the message being published.

        """
        if topic not in self.available_topics:
            raise ValueError(
                f"Topic {topic} is not available within 1.0 seconds. Check if the topic exists."
            )
        if message_type not in self.available_message_types:
            raise TypeError(
                "Expected message one of message types: {}, got {}".format(
                    self.available_message_types, message_type
                )
            )

        model = self.available_topic_models[message_type]
        try:
            model.model_validate(message)
        except ValidationError as e:
            raise ValueError(f"Failed to populate fields: {e}")

        return "Message published successfully"


class MockGetROS2MessageInterfaceTool(GetROS2MessageInterfaceTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    mock_interfaces: Dict[str, str]

    def _run(self, msg_type: str) -> str:
        """
        Mocked method that returns the interface definition for a given ROS2 message type.

        Parameters
        ----------
        msg_type : str
            The ROS2 message type for which to retrieve the interface definition.

        Returns
        -------
        str
            The mocked output of 'ros2 interface show' for the specified message type.
        """
        if msg_type in self.mock_interfaces:
            return self.mock_interfaces[msg_type]
        else:
            raise ImportError(f"Module {msg_type} not found.")


class MockCallROS2ServiceTool(CallROS2ServiceTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_services: List[str]
    available_service_types: List[str]
    available_service_models: Dict[str, Type[BaseModel]]

    def _run(
        self,
        service_name: str,
        service_type: str,
        service_args: Dict[str, Any],
    ) -> str:
        if service_name not in self.available_services:
            raise ValueError(
                f"Service {service_name} is not available within 1.0 seconds. Check if the service exists."
            )
        if service_type not in self.available_service_types:
            raise TypeError(
                "Expected one of service types: {}, got {}".format(
                    self.available_service_types, service_type
                )
            )
        if service_type in self.available_service_models:
            model = self.available_service_models[service_type]
            try:
                model.model_validate(service_args)
            except ValidationError as e:
                raise ValueError(f"Failed to populate fields: {e}")
            response = ROS2Message(payload={"response": "success"})
            return str(
                {
                    "payload": response.payload,
                    "metadata": response.metadata,
                }
            )
        else:
            raise KeyError(
                f"Model for service type {service_type} not included in models"
            )


class MockGetROS2ServicesNamesAndTypesTool(GetROS2ServicesNamesAndTypesTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    mock_service_names_and_types: list[str]

    def _run(self) -> str:
        """Mocked method that returns the mock topics and types instead of fetching from ROS2.

        Returns
        -------
        str
            Mocked output of 'get_ros2_topics_names_and_types' tool.
        """
        return "\n".join(self.mock_service_names_and_types)


class MockGetROS2ActionsNamesAndTypesTool(GetROS2ActionsNamesAndTypesTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    mock_actions_names_and_types: list[str]

    def _run(self) -> str:
        """Mocked method that returns the mock topics and types instead of fetching from ROS2.

        Returns
        -------
        str
            Mocked output of 'get_ros2_topics_names_and_types' tool.
        """
        return "\n".join(self.mock_actions_names_and_types)


class MockStartROS2ActionTool(StartROS2ActionTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_actions: List[str] = []
    available_action_types: List[str] = []
    available_action_models: Dict[str, Type[BaseModel]]

    def _run(
        self, action_name: str, action_type: str, action_args: Dict[str, Any]
    ) -> str:
        if action_name not in self.available_actions:
            raise ValueError(
                f"Action {action_name} is not available within 1.0 seconds. Check if the action exists."
            )
        if action_type not in self.available_action_types:
            raise TypeError(
                f"Expected one of action types: {self.available_action_types}, got {action_type}"
            )
        model = self.available_action_models[action_type]
        try:
            model.model_validate(action_args)
        except ValidationError as e:
            raise ValueError(f"Failed to populate fields: {e}")

        action_id = str(uuid.uuid4())
        response = action_id
        self.internal_action_id_mapping[response] = action_id
        return "Action started with ID: " + response


class MockCancelROS2ActionTool(CancelROS2ActionTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_action_ids: List[str] = []

    def _run(self, action_id: str) -> str:
        if action_id not in self.available_action_ids:
            raise ValueError(f"Action {action_id} is not available for cancellation.")
        return f"Action {action_id} cancelled"


class MockGetROS2ActionFeedbackTool(GetROS2ActionFeedbackTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    available_feedbacks: Dict[str, List[Any]] = {}
    internal_action_id_mapping: Dict[str, str] = {}
    action_feedbacks_store_lock: Lock = Lock()

    def _run(self, action_id: str) -> str:
        if action_id not in self.internal_action_id_mapping:
            raise KeyError(f"Action ID {action_id} not found in internal mapping.")
        external_id = self.internal_action_id_mapping[action_id]
        with self.action_feedbacks_store_lock:
            feedbacks = self.available_feedbacks.get(external_id, [])
            self.available_feedbacks[external_id] = []
        return str(feedbacks)


class MockGetROS2ActionResultTool(GetROS2ActionResultTool):
    available_results: Dict[str, Any] = {}
    internal_action_id_mapping: Dict[str, str] = {}
    action_results_store_lock: Lock = Lock()

    def _run(self, action_id: str) -> str:
        if action_id not in self.internal_action_id_mapping:
            raise KeyError(f"Action ID {action_id} not found in internal mapping.")
        external_id = self.internal_action_id_mapping[action_id]
        with self.action_results_store_lock:
            if external_id not in self.available_results:
                raise ValueError(f"No result available for action {action_id}")
            result = self.available_results[external_id]
        return str(result)


class MockGetROS2ActionIDsTool(GetROS2ActionIDsTool):
    internal_action_id_mapping: Dict[str, str] = {}

    def _run(self) -> str:
        return str(list(self.internal_action_id_mapping.keys()))


class MockActionsToolkit(ROS2ActionToolkit):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    mock_actions_names_and_types: list[str] = []
    available_actions: List[str] = []
    available_action_types: List[str] = []
    available_action_models: Dict[str, Type[BaseModel]] = {}

    def get_tools(self) -> List[BaseTool]:
        return [
            MockStartROS2ActionTool(
                available_actions=self.available_actions,
                feedback_callback=self._generic_feedback_callback,
                on_done_callback=self._generic_on_done_callback,
                available_action_types=self.available_action_types,
                available_action_models=self.available_action_models,
            ),
            MockCancelROS2ActionTool(),
            MockGetROS2ActionFeedbackTool(),
            MockGetROS2ActionResultTool(),
            MockGetROS2ActionIDsTool(),
            MockGetROS2ActionsNamesAndTypesTool(
                mock_actions_names_and_types=self.mock_actions_names_and_types
            ),
        ]


class MockGetDistanceToObjectsTool(GetDistanceToObjectsTool):
    connector: ROS2Connector = MagicMock(spec=ROS2Connector)
    node: MagicMock = MagicMock()
    mock_distance_measurements: List[DistanceMeasurement] = []
    available_topics: List[str]

    def _run(self, camera_topic: str, depth_topic: str, object_names: list[str]):
        """Method that returns a mock message with the distance to the objects.

        Parameters
        ----------
        camera_topic : str
            Topic with the camera image
        depth_topic : str
            Topic with the depth image
        object_names : list[str]
            List of object names to get the distance to

        Returns
        -------
        str
            Message from the tool
        """
        if camera_topic not in self.available_topics:
            return f"Topic {camera_topic} is not available within 1.0 seconds. Check if the topic exists."
        if depth_topic not in self.available_topics:
            return f"Topic {depth_topic} is not available within 1.0 seconds. Check if the topic exists."
        measurement_string = ", ".join(
            [
                f"{measurement[0]}: {measurement[1]:.2f}m away"
                for measurement in self.mock_distance_measurements
                if measurement[0] in object_names
            ]
        )

        return f"I have detected the following items in the picture {measurement_string or 'no objects'}"



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import streamlit as st

from rai_bench.results_processing.data_loading import (
    BenchmarkResults,
    ModelResults,
    RunResults,
    get_available_runs,
    load_run_results,
)
from rai_bench.results_processing.data_processing import (
    analyze_subtasks,
    create_extra_calls_dataframe,
    create_model_summary_dataframe,
    create_task_details_dataframe,
    create_task_metrics_dataframe,
    get_available_extra_tool_calls,
    get_unique_values_from_results,
)
from rai_bench.tool_calling_agent.results_tracking import (
    TaskResult,
)

EXPERIMENT_DIR = "./src/rai_bench/rai_bench/experiments"
DETAILED_FILE_NAME: str = "results.csv"
SUMMARY_FILE_NAME: str = "results_summary.csv"


def adjust_bar_width(
    fig: go.Figure,
    max_full_width_bars: int = 10,
    base_width: float = 0.8,
    bargap: float = 0.1,
) -> go.Figure:
    """
    Adjust bar width dynamically based on number of bars in the figure.

    Parameters
    ----------
    fig : go.Figure
        A Plotly figure containing one or more bar traces.
    max_full_width_bars : int, optional
        Number of bars to display at full base_width before scaling kicks in.
    base_width : float, optional
        Width of each bar (as a fraction of its category slot) when few bars.
    bargap : float, optional
        Fractional gap between bars (0 to 1).

    """
    try:
        first_bar = next(trace for trace in fig.data if trace.type == "bar")
        n_bars = len(first_bar.x)
    except StopIteration:
        return fig

    scale = min(n_bars, max_full_width_bars) / max_full_width_bars
    width = base_width * scale
    fig.update_traces(selector={"type": "bar"}, width=width, offset=0)  # type: ignore
    fig.update_layout(xaxis_type="category", bargap=bargap)  # type: ignore
    return fig


def create_bar_chart(
    df: pd.DataFrame,
    x_column: str,
    y_column: str,
    title: str,
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    color_column: Optional[str] = None,
    custom_data: Optional[List[str]] = None,
    hover_template: Optional[str] = None,
    y_range: Optional[Tuple[float, float]] = None,
    x_tickvals: Optional[List[Any]] = None,
    x_ticktext: Optional[List[str]] = None,
) -> go.Figure:
    """
    Create a standardized bar chart with consistent styling.
    """
    # Set default labels if not provided
    if x_label is None:
        x_label = x_column
    if y_label is None:
        y_label = y_column

    # Create labels dictionary
    labels = {x_column: x_label, y_column: y_label}

    # Create the chart
    fig = px.bar(
        df,
        x=x_column,
        y=y_column,
        title=title,
        labels=labels,
        color=color_column,
        barmode="group" if color_column else "relative",
        custom_data=custom_data,
    )

    # Apply common styling
    fig.update_layout(xaxis_tickangle=-45)  # type: ignore

    # Apply optional customizations
    if hover_template:
        fig.update_traces(hovertemplate=hover_template)  # type: ignore

    if y_range:
        fig.update_yaxes(range=y_range)  # type: ignore

    if x_tickvals and x_ticktext:
        fig.update_xaxes(tickvals=x_tickvals, ticktext=x_ticktext)  # type: ignore

    return adjust_bar_width(fig=fig)


def create_aggregate_model_metrics(run_results: RunResults) -> pd.DataFrame:
    """
    Create a dataframe with aggregated metrics for each model across all benchmarks.

    Parameters
    ----------
    run_results : RunResults
        The RunResults object containing all benchmark data

    Returns
    -------
    pd.DataFrame
        A dataframe with aggregated metrics for each model
    """
    # Dictionary to store aggregated metrics for each model
    model_metrics: Dict[str, Dict[str, Any]] = {}

    # Track total counts for averaging
    model_counts: Dict[str, int] = {}

    # Collect metrics from all benchmarks
    for benchmark_name, benchmark in run_results.benchmarks.items():
        for model_name, model_results in benchmark.models.items():
            if not model_results.check_if_summaries_present():
                continue

            if model_name not in model_metrics:
                model_metrics[model_name] = {
                    "success_rate_sum": 0.0,
                    "time_sum": 0.0,
                    "extra_tool_calls_sum": 0.0,
                    "total_extra_tool_calls_used": 0,
                    "count": 0,
                }
                model_counts[model_name] = 0

            # Sum up metrics for this model from this benchmark
            metrics = model_metrics[model_name]
            metrics["success_rate_sum"] += (
                model_results.avg_success_rate * model_results.count
            )
            metrics["time_sum"] += model_results.avg_time * model_results.count
            metrics["extra_tool_calls_sum"] += (
                model_results.avg_extra_tool_calls * model_results.count
            )

            # Sum up total extra tool calls used (not averaged)
            for run in model_results.runs:
                for summary in run.benchmark_summaries:
                    metrics["total_extra_tool_calls_used"] += (
                        summary.total_extra_tool_calls_used
                    )

            # Add to the count for this model
            metrics["count"] += model_results.count
            model_counts[model_name] += model_results.count

    # Convert to dataframe and calculate averages
    result_data = []
    for model_name, metrics in model_metrics.items():
        count = metrics["count"]
        if count > 0:
            result_data.append(
                {
                    "model_name": model_name,
                    "avg_success_rate": (metrics["success_rate_sum"] / count),
                    "avg_time": metrics["time_sum"] / count,
                    "avg_extra_tool_calls": metrics["extra_tool_calls_sum"] / count,
                    "total_extra_tool_calls_used": metrics[
                        "total_extra_tool_calls_used"
                    ],
                    "total_tasks": count,
                }
            )

    return pd.DataFrame(result_data)


def display_aggregate_model_metrics(df: pd.DataFrame):
    """Display aggregate model metrics across all benchmarks."""
    if df.empty:
        st.warning("No aggregate data available for visualization.")
        return

    fig1 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_success_rate",
        title="Average Success Rate by Model (Across All Benchmarks)",
        x_label="Model Name",
        y_label="Success Rate (%)",
        color_column="model_name",
        y_range=(0.0, 100.0),
    )
    st.plotly_chart(fig1, use_container_width=True)  # type: ignore

    fig2 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_time",
        title="Average Completion Time by Model (Across All Benchmarks)",
        x_label="Model Name",
        y_label="Avg Time (seconds)",
        color_column="model_name",
    )
    st.plotly_chart(fig2, use_container_width=True)  # type: ignore

    fig3 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_extra_tool_calls",
        title="Average Extra Tool Calls by Model (Across All Benchmarks)",
        x_label="Model Name",
        y_label="Avg Extra Tool Calls",
        color_column="model_name",
    )
    st.plotly_chart(fig3, use_container_width=True)  # type: ignore


def render_overall_performance(run_results: RunResults):
    """Render the overall performance summary across all benchmarks."""
    aggregate_df = create_aggregate_model_metrics(run_results)

    if aggregate_df.empty:
        st.warning("No aggregate data available.")
    else:
        st.subheader("Aggregate Performance Metrics")
        st.write(
            "These charts show average metrics across all benchmarks for each model"
        )
        display_aggregate_model_metrics(aggregate_df)


def display_models_summ_data(df: pd.DataFrame):
    """Display summary data charts for models."""
    if df.empty:
        st.warning("No summary data available for visualization.")
        return

    fig1 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_success_rate",
        title="Success Rate by Model",
        x_label="Model Name",
        y_label="Success Rate (%)",
        color_column="model_name",
        y_range=(0.0, 100.0),
    )
    st.plotly_chart(fig1, use_container_width=True)  # type: ignore

    fig2 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_time",
        title="Average Completion Time by Model",
        x_label="Model Name",
        y_label="Avg Time (seconds)",
        color_column="model_name",
    )
    st.plotly_chart(fig2, use_container_width=True)  # type: ignore

    fig3 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="total_extra_tool_calls_used",
        title="Total Extra Tool Calls Used by Model",
        x_label="Model Name",
        y_label="Total Extra Tool Calls",
        color_column="model_name",
    )
    st.plotly_chart(fig3, use_container_width=True)  # type: ignore


def display_models_extra_calls_data(df: pd.DataFrame, extra_calls: int):
    """Display data for models with specific extra tool calls."""
    if df.empty:
        st.warning(f"No data available for {extra_calls} extra tool calls.")
        return

    fig1 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_success_rate",
        title=f"Success Rate by Model (Extra Tool Calls: {extra_calls})",
        x_label="Model Name",
        y_label="Success Rate (%)",
        color_column="model_name",
        y_range=(0.0, 1.0),
    )
    st.plotly_chart(fig1, use_container_width=True)  # type: ignore

    fig2 = create_bar_chart(
        df=df,
        x_column="model_name",
        y_column="avg_time",
        title=f"Avg Completion Time by Model (Extra Tool Calls: {extra_calls})",
        x_label="Model Name",
        y_label="Avg Time (s)",
        color_column="model_name",
    )
    st.plotly_chart(fig2, use_container_width=True)  # type: ignore


def display_task_type_performance(model_results: ModelResults):
    """Display performance charts by task type."""
    task_type_df = create_task_metrics_dataframe(model_results, "type")

    if task_type_df.empty:
        st.warning("No task type data available.")
        return

    fig_type_score = create_bar_chart(
        df=task_type_df,
        x_column="type",
        y_column="avg_score",
        title="Success Rate by Task Type",
        x_label="Task Type",
        y_label="Avg Score",
        y_range=(0.0, 1.0),
    )
    st.plotly_chart(fig_type_score, use_container_width=True)  # type: ignore

    fig_type_calls = create_bar_chart(
        df=task_type_df,
        x_column="type",
        y_column="avg_extra_tool_calls",
        title="Avg Extra Tool Calls Used by Task Type",
        x_label="Task Type",
        y_label="Avg Extra Tool Calls Used",
    )
    st.plotly_chart(fig_type_calls, use_container_width=True)  # type: ignore


def display_task_complexity_performance(model_results: ModelResults):
    """Display performance charts by task complexity."""
    complexity_df = create_task_metrics_dataframe(model_results, "complexity")

    if complexity_df.empty:
        st.warning("No complexity data available.")
        return

    fig_complexity_score = create_bar_chart(
        df=complexity_df,
        x_column="complexity",
        y_column="avg_score",
        title="Success Rate by Task Complexity",
        x_label="Task Complexity",
        y_label="Avg Score",
        y_range=(0.0, 1.0),
    )
    st.plotly_chart(fig_complexity_score, use_container_width=True)  # type: ignore

    fig_complexity_calls = create_bar_chart(
        df=complexity_df,
        x_column="complexity",
        y_column="avg_extra_tool_calls",
        title="Avg Extra Tool Calls Used by Task Complexity",
        x_label="Task Complexity",
        y_label="Avg Extra Tool Calls Used",
    )
    st.plotly_chart(fig_complexity_calls, use_container_width=True)  # type: ignore


def wrap_text(text: str, max_width: int = 50):
    """Wrap text to multiple lines if it's too long"""
    words = text.split()
    lines: List[str] = []
    current_line = []
    current_length = 0

    for word in words:
        if current_length + len(word) + len(current_line) <= max_width:
            current_line.append(word)
            current_length += len(word)
        else:
            lines.append(" ".join(current_line))
            current_line = [word]
            current_length = len(word)

    if current_line:
        lines.append(" ".join(current_line))

    return "<br>".join(lines)


def display_detailed_task_type_analysis(
    model_results: ModelResults, selected_type: str
):
    """Display detailed analysis for a specific task type."""
    # first, get only the tasks of the selected type
    tasks_for_type_df = create_task_details_dataframe(model_results, selected_type)
    if tasks_for_type_df.empty:
        st.warning(f"No tasks of type {selected_type} found.")
        return
    # Now aggregate by complexity for that type
    filtered_by_complexity = (
        tasks_for_type_df.groupby("complexity")  # type: ignore
        .agg(
            avg_score=("score", "mean"),
            avg_time=("total_time", "mean"),
            avg_extra_tool_calls=("extra_tool_calls_used", "mean"),
        )
        .reset_index()
    )
    filtered_by_complexity = filtered_by_complexity[
        filtered_by_complexity["complexity"].notna()
    ]

    # Display success rate by complexity for the selected task type
    if not filtered_by_complexity.empty:
        fig_complexity_score = create_bar_chart(
            df=filtered_by_complexity,
            x_column="complexity",
            y_column="avg_score",
            title=f"Success Rate by Task Complexity for '{selected_type}' Tasks",
            x_label="Task Complexity",
            y_label="Avg Score",
            y_range=(0.0, 1.0),
        )
        st.plotly_chart(fig_complexity_score, use_container_width=True)  # type: ignore

    # Display success rate by individual task
    task_details_df = create_task_details_dataframe(model_results, selected_type)

    if task_details_df.empty:
        st.warning(f"No task details available for type: {selected_type}")
        return

    task_stats = (
        task_details_df.groupby("task_prompt")  # type: ignore
        .agg({"score": "mean", "total_time": "mean"})
        .reset_index()
    )

    # Replace zero scores with a small value to enable hover
    task_stats = task_stats.replace(0, 0.01)  # type: ignore

    # wrapped text column for hover, so that text is multiline
    task_stats["wrapped_prompt"] = task_stats["task_prompt"].apply(wrap_text)  # type: ignore
    task_stats["score"] = task_stats["score"].round(2)
    task_stats["total_time"] = task_stats["total_time"].round(2)
    # Create short labels for x-axis
    short_labels: List[str] = [
        (t[:30] + "...") if len(t) > 30 else t for t in task_stats["task_prompt"]
    ]
    st.warning(
        "The 0.01 values corresponds to 0. They were set to 0.01 to enable hovering over them so you can see the whole task prompt"
    )
    fig_task = create_bar_chart(
        df=task_stats,
        x_column="task_prompt",
        y_column="score",
        title=f"Avg Score for '{selected_type}' Tasks",
        x_label="Task",
        y_label="Avg Score",
        custom_data=["wrapped_prompt", "score"],
        hover_template="<b>Task:</b> %{customdata[0]}\n <b>Score:</b> %{customdata[1]}",
        y_range=(0.0, 1.0),
        x_tickvals=task_stats["task_prompt"].tolist(),
        x_ticktext=short_labels,
    )
    st.plotly_chart(fig_task, use_container_width=True)  # type: ignore

    fig_time = create_bar_chart(
        df=task_stats,
        x_column="task_prompt",
        y_column="total_time",
        title=f"Avg Time for '{selected_type}' Tasks",
        x_label="Task",
        y_label="Avg Time (s)",
        custom_data=["wrapped_prompt", "total_time"],
        hover_template="<b>Task:</b> %{customdata[0]}\n <b>Time:</b> %{customdata[1]}",
        x_tickvals=task_stats["task_prompt"].tolist(),
        x_ticktext=short_labels,
    )
    st.plotly_chart(fig_time, use_container_width=True)  # type: ignore


def display_subtask_analysis(
    subtask_df: pd.DataFrame, tool_name_df: pd.DataFrame, error_df: pd.DataFrame
):
    """Display subtask analysis tables."""
    # Subtasks table
    st.subheader("Subtasks")
    st.dataframe(subtask_df, use_container_width=True)  # type: ignore

    # Tool name summary
    st.subheader("Subtasks grouped by tool name")
    st.dataframe(tool_name_df, use_container_width=True)  # type: ignore

    # Error summary
    st.subheader("Errors by subtask")
    st.dataframe(error_df, use_container_width=True)  # type: ignore

    # Additional analysis: group errors by tool name
    if not error_df.empty:
        error_by_tool_df = (
            error_df.groupby("expected_tool_name")  # type: ignore
            .agg(
                errors_count=("count", "sum"),
            )
            .reset_index()
            .sort_values("errors_count", ascending=False)
        )

        st.subheader("Errors grouped by tool name")
        st.dataframe(error_by_tool_df, use_container_width=True)  # type: ignore


def display_validator_results(task_result: TaskResult):
    """Display validation results for a single task."""
    col1, col2 = st.columns(2)

    # Task overview in the first column
    col1.markdown("#### Task Overview")
    col1.markdown(f"**Type:** {task_result.type}")
    col1.markdown(f"**Complexity:** {task_result.complexity}")
    col1.markdown(f"**Score:** {task_result.score}")
    col1.markdown(f"**Time:** {task_result.total_time:.2f}s")
    col1.markdown(f"**Extra Tool Calls Used:** {task_result.extra_tool_calls_used}")

    # Validation details in the second column
    col2.markdown("#### Validation Results")

    if not task_result.validation_info:
        col2.warning("No validation information available.")
    else:
        for idx, validator in enumerate(task_result.validation_info):
            col2.markdown(f"**Validator {idx + 1}:** {validator.type}")
            col2.markdown(f"**Passed:** {'✅' if validator.passed else '❌'}")

            # Subtasks table
            if validator.subtasks:
                subtask_data = []
                for subtask_idx, subtask in enumerate(validator.subtasks):
                    expected_tool = subtask.args.get("expected_tool_name", "Unknown")
                    subtask_data.append(
                        {
                            "Subtask": subtask_idx + 1,
                            "Expected Tool": expected_tool,
                            "Passed": "✅" if subtask.passed else "❌",
                            "Errors": (
                                ", ".join(subtask.errors) if subtask.errors else "None"
                            ),
                        }
                    )

                col2.dataframe(pd.DataFrame(subtask_data), use_container_width=True)
            else:
                col2.info("No subtasks in this validator.")


def render_model_performance_tab(bench_results: BenchmarkResults):
    st.header("Model Performance")

    summ_df = create_model_summary_dataframe(bench_results)
    display_models_summ_data(summ_df)

    # Extra tool calls analysis
    st.subheader("Performance by Extra Tool Calls")
    extra_calls_values = get_available_extra_tool_calls(bench_results)

    if not extra_calls_values:
        st.warning("No extra tool calls data available.")
        return

    extra_tool_calls_selected = st.radio(
        "Maximum extra tool calls:", extra_calls_values, key="extra_tool_calls"
    )

    extra_calls_df = create_extra_calls_dataframe(
        benchmark=bench_results, extra_calls=extra_tool_calls_selected
    )

    display_models_extra_calls_data(
        df=extra_calls_df, extra_calls=extra_tool_calls_selected
    )


def render_task_performance_tab(bench_results: BenchmarkResults):
    """Render the task performance across models tab."""
    st.header("Performance Across Tasks")

    model_names = list(bench_results.models.keys())
    if not model_names:
        st.warning(f"No models available for benchmark: {bench_results.benchmark_name}")
        return

    model_selected = st.radio("Model:", model_names, key="task_model")
    model_results = bench_results.models[model_selected]

    st.subheader(f"Model: {model_selected}")
    st.write(f"Data aggregated across {len(model_results.runs)} runs")

    # Display performance by task type
    st.subheader("Performance by Task Type")
    display_task_type_performance(model_results)

    # Display performance by complexity
    st.subheader("Performance by Task Complexity")
    display_task_complexity_performance(model_results)

    # Per Task Type Analysis
    st.subheader("Detailed Task Type Analysis")
    task_types = get_unique_values_from_results(model_results, "type")

    if not task_types:
        st.warning("No task types available.")
        return

    selected_type = st.selectbox(
        "Select Task Type", sorted(task_types), key="task_type"
    )
    display_detailed_task_type_analysis(model_results, selected_type)


def render_validator_analysis_tab(bench_results: BenchmarkResults):
    """Render the validator analysis tab."""
    st.header("Validator Analysis")
    st.info("This tab provides detailed information about validation results.")

    model_names = list(bench_results.models.keys())
    if not model_names:
        st.warning(f"No models available for benchmark: {bench_results.benchmark_name}")
        return

    model_selected = st.selectbox("Model", model_names, key="val_model")
    model_results = bench_results.models[model_selected]

    # Run selection (if there are multiple runs)
    if len(model_results.runs) > 1:
        run_ids = [f"Run {run.run_id}" for run in model_results.runs]
        run_selected_idx = st.selectbox(
            "Run", range(len(run_ids)), format_func=lambda x: run_ids[x]
        )
        run = model_results.runs[run_selected_idx]
    else:
        run = model_results.runs[0]

    st.subheader("Per-Task Validators & Subtasks")

    # Display detailed validation results for each task
    for task_result in run.task_results:
        with st.expander(f"Task: {task_result.task_prompt}"):
            display_validator_results(task_result)


def render_subtask_analysis_tab(bench_results: BenchmarkResults):
    """Render the subtask analysis tab."""
    st.header("Subtask Analysis")

    # Model selection
    model_names = list(bench_results.models.keys())
    if not model_names:
        st.warning(f"No models available for benchmark: {bench_results.benchmark_name}")
        return

    model_selected = st.selectbox("Model", model_names, key="sub_model")

    # Analyze subtasks
    with st.spinner("Analyzing subtasks..."):
        subtask_df, tool_name_df, error_df = analyze_subtasks(
            bench_results, model_selected
        )

    # Display results
    if subtask_df.empty:
        st.warning("No subtask data available for analysis.")
        return

    # Subtasks table
    st.subheader("Subtasks")
    st.dataframe(subtask_df, use_container_width=True)  # type:ignore

    # Tool name summary
    st.subheader("Subtasks grouped by tool name")
    st.dataframe(tool_name_df, use_container_width=True)  # type:ignore

    # Error summary
    st.subheader("Errors by subtask")
    st.dataframe(error_df, use_container_width=True)  # type:ignore

    # Additional analysis: group errors by tool name
    if not error_df.empty:
        error_by_tool_df = (
            error_df.groupby("expected_tool_name")  # type:ignore
            .agg(
                errors_count=("count", "sum"),
            )
            .reset_index()
            .sort_values("errors_count", ascending=False)
        )

        st.subheader("Errors grouped by tool name")
        st.dataframe(error_by_tool_df, use_container_width=True)  # type:ignore


def render_manipulation_o3de(bench_results: BenchmarkResults): ...


def render_tool_calling_agent(bench_results: BenchmarkResults):
    tabs = st.tabs(
        [
            "Model Performance",
            "Task Performance",
            "Validator Analysis",
            "Subtask Analysis",
        ]
    )

    with tabs[0]:
        render_model_performance_tab(bench_results)

    with tabs[1]:
        render_task_performance_tab(bench_results)

    with tabs[2]:
        render_validator_analysis_tab(bench_results)

    with tabs[3]:
        render_subtask_analysis_tab(bench_results)


BENCHMARK_SECTIONS: Dict[str, Any] = {
    "manipulation_o3de": render_manipulation_o3de,
    "tool_calling_agent": render_tool_calling_agent,
}


def main():
    st.set_page_config(layout="wide", page_title="LLM Task Results Visualizer")
    st.title("RAI BENCHMARK RESULTS")

    run_folders = get_available_runs(EXPERIMENT_DIR)

    if not run_folders:
        st.warning("No benchmark runs found in the experiments directory.")
        return

    if "run_results" not in st.session_state:
        st.session_state.run_results = None
    if "selected_benchmark" not in st.session_state:
        st.session_state.selected_benchmark = None

    selected = st.selectbox("Select run folder", run_folders)
    if st.button("Load Run Data") or st.session_state.run_results is None:
        results_dir = os.path.join(EXPERIMENT_DIR, selected)
        with st.spinner("Loading run data..."):
            st.session_state.run_results = load_run_results(results_dir)

    run_results: Optional[RunResults] = st.session_state.run_results

    if run_results is None:
        st.error("No valid benchmark data found in the selected run.")
        return

    # Create tabs
    st.markdown("---")
    st.header("Overview")

    # Display overall performance summary across all benchmarks
    render_overall_performance(run_results)

    # Benchmark selection
    st.markdown("---")
    st.header("Benchmark-Specific Analysis")

    benchmark_names = list(run_results.benchmarks.keys())
    if not benchmark_names:
        st.warning("No benchmarks available in this run.")
        return

    selected_benchmark_name = st.selectbox(
        "Select a benchmark for detailed analysis:",
        benchmark_names,
        index=(
            benchmark_names.index(st.session_state.selected_benchmark)
            if st.session_state.selected_benchmark in benchmark_names
            else 0
        ),
    )

    # Store the selected benchmark in session state
    st.session_state.selected_benchmark = selected_benchmark_name
    selected_benchmark = run_results.benchmarks[selected_benchmark_name]

    BENCHMARK_SECTIONS[selected_benchmark_name](selected_benchmark)


if __name__ == "__main__":
    main()



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, List, Type

from langchain_core.tools import BaseTool
from pydantic import BaseModel
from rai_open_set_vision.tools.gdino_tools import (
    DistanceMeasurement,
)

from rai_bench.tool_calling_agent.interfaces import Task
from rai_bench.tool_calling_agent.messages.actions import (
    NavigateToPoseGoal,
    SpinGoal,
)
from rai_bench.tool_calling_agent.mocked_tools import (
    MockActionsToolkit,
    MockGetDistanceToObjectsTool,
    MockGetROS2ActionFeedbackTool,
    MockGetROS2ActionResultTool,
    MockGetROS2ActionsNamesAndTypesTool,
    MockGetROS2MessageInterfaceTool,
    MockGetROS2TopicsNamesAndTypesTool,
    MockStartROS2ActionTool,
)

loggers_type = logging.Logger

ROBOT_NAVIGATION_SYSTEM_PROMPT = """You are an autonomous robot connected to ros2 environment. Your main goal is to fulfill the user's requests.
    Do not make assumptions about the environment you are currently in.
    You can use ros2 topics, services and actions to operate.

    <rule> As a first step check transforms by getting 1 message from /tf topic </rule>
    <rule> use /cmd_vel topic very carefully. Obstacle detection works only with nav2 stack, so be careful when it is not used. </rule>>
    <rule> be patient with running ros2 actions. usually the take some time to run. </rule>
    <rule> Always check your transform before and after you perform ros2 actions, so that you can verify if it worked. </rule>

    Navigation tips:
    - it's good to start finding objects by rotating, then navigating to some diverse location with occasional rotations. Remember to frequency detect objects.
    - for driving forward/backward or to some coordinates, ros2 actions are better.
    - for driving for some specific time or in specific manner (like shaper or turns) it good to use /cmd_vel topic
    - you are currently unable to read map or point-cloud, so please avoid subscribing to such topics.
    - if you are asked to drive towards some object, it's good to:
        1. check the camera image and verify if objects can be seen
        2. if only driving forward is required, do it
        3. if obstacle avoidance might be required, use ros2 actions navigate_*, but first check your current position, then very accurately estimate the goal pose.
    - it is good to verify using given information if the robot is not stuck
    - navigation actions sometimes fail. Their output can be read from rosout. You can also tell if they partially worked by checking the robot position and rotation.
    - before using any ros2 interfaces, always make sure to check you are using the right interface
    - processing camera image takes 5-10s. Take it into account that if the robot is moving, the information can be outdated. Handle it by good planning of your movements.
    - you are encouraged to use wait tool in between checking the status of actions
    - to find some object navigate around and check the surrounding area
    - when the goal is accomplished please make sure to cancel running actions
    - when you reach the navigation goal - double check if you reached it by checking the current position
    - if you detect collision, please stop operation
    - you will be given your camera image description. Based on this information you can reason about positions of objects.
    - be careful and aboid obstacles

    Here are the corners of your environment:
    (-2.76,9.04, 0.0),
    (4.62, 9.07, 0.0),
    (-2.79, -3.83, 0.0),
    (4.59, -3.81, 0.0)

    This is location of places:
    Kitchen:
    (2.06, -0.23, 0.0),
    (2.07, -1.43, 0.0),
    (-2.44, -0.38, 0.0),
    (-2.56, -1.47, 0.0)

    # Living room:
    (-2.49, 1.87, 0.0),
    (-2.50, 5.49, 0.0),
    (0.79, 5.73, 0.0),
    (0.92, 1.01, 0.0)

    Before starting anything, make sure to load available topics, services and actions.
    Example tool calls:
    - get_ros2_message_interface, args: {'msg_type': 'turtlesim/srv/TeleportAbsolute'}
    - publish_ros2_message, args: {'topic': '/cmd_vel', 'message_type': 'geometry_msgs/msg/Twist', 'message': {linear: {x: 0.5, y: 0.0, z: 0.0}, angular: {x: 0.0, y: 0.0, z: 1.0}}}
    - start_ros2_action, args: {'action_name': '/dock', 'action_type': 'nav2_msgs/action/Dock', 'action_args': {}}
    """

TOPICS_NAMES_AND_TYPES = [
    "topic: /assisted_teleop/_action/feedback\ntype: nav2_msgs/action/AssistedTeleop_FeedbackMessage\n",
    "topic: /assisted_teleop/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /backup/_action/feedback\ntype: nav2_msgs/action/BackUp_FeedbackMessage\n",
    "topic: /backup/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /behavior_server/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /behavior_tree_log\ntype: nav2_msgs/msg/BehaviorTreeLog\n",
    "topic: /bond\ntype: bond/msg/Status\n",
    "topic: /bt_navigator/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /camera/camera/color/camera_info\ntype: sensor_msgs/msg/CameraInfo\n",
    "topic: /camera/camera/color/image_raw\ntype: sensor_msgs/msg/Image\n",
    "topic: /camera/camera/depth/camera_info\ntype: sensor_msgs/msg/CameraInfo\n",
    "topic: /camera/camera/depth/image_rect_raw\ntype: sensor_msgs/msg/Image\n",
    "topic: /clock\ntype: rosgraph_msgs/msg/Clock\n",
    "topic: /cmd_vel_nav\ntype: geometry_msgs/msg/Twist\n",
    "topic: /cmd_vel_teleop\ntype: geometry_msgs/msg/Twist\n",
    "topic: /compute_path_through_poses/_action/feedback\ntype: nav2_msgs/action/ComputePathThroughPoses_FeedbackMessage\n",
    "topic: /compute_path_through_poses/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /compute_path_to_pose/_action/feedback\ntype: nav2_msgs/action/ComputePathToPose_FeedbackMessage\n",
    "topic: /compute_path_to_pose/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /controller_server/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /diagnostics\ntype: diagnostic_msgs/msg/DiagnosticArray\n",
    "topic: /drive_on_heading/_action/feedback\ntype: nav2_msgs/action/DriveOnHeading_FeedbackMessage\n",
    "topic: /drive_on_heading/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /follow_path/_action/feedback\ntype: nav2_msgs/action/FollowPath_FeedbackMessage\n",
    "topic: /follow_path/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /follow_waypoints/_action/feedback\ntype: nav2_msgs/action/FollowWaypoints_FeedbackMessage\n",
    "topic: /follow_waypoints/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /global_costmap/costmap\ntype: nav_msgs/msg/OccupancyGrid\n",
    "topic: /global_costmap/costmap_raw\ntype: nav2_msgs/msg/Costmap\n",
    "topic: /global_costmap/costmap_updates\ntype: map_msgs/msg/OccupancyGridUpdate\n",
    "topic: /global_costmap/footprint\ntype: geometry_msgs/msg/Polygon\n",
    "topic: /global_costmap/global_costmap/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /global_costmap/published_footprint\ntype: geometry_msgs/msg/PolygonStamped\n",
    "topic: /global_costmap/scan\ntype: sensor_msgs/msg/LaserScan\n",
    "topic: /goal_pose\ntype: geometry_msgs/msg/PoseStamped\n",
    "topic: /led_strip\ntype: sensor_msgs/msg/Image\n",
    "topic: /local_costmap/costmap\ntype: nav_msgs/msg/OccupancyGrid\n",
    "topic: /local_costmap/costmap_raw\ntype: nav2_msgs/msg/Costmap\n",
    "topic: /local_costmap/costmap_updates\ntype: map_msgs/msg/OccupancyGridUpdate\n",
    "topic: /local_costmap/footprint\ntype: geometry_msgs/msg/Polygon\n",
    "topic: /local_costmap/local_costmap/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /local_costmap/published_footprint\ntype: geometry_msgs/msg/PolygonStamped\n",
    "topic: /local_costmap/scan\ntype: sensor_msgs/msg/LaserScan\n",
    "topic: /map\ntype: nav_msgs/msg/OccupancyGrid\n",
    "topic: /map_metadata\ntype: nav_msgs/msg/MapMetaData\n",
    "topic: /map_saver/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /navigate_through_poses/_action/feedback\ntype: nav2_msgs/action/NavigateThroughPoses_FeedbackMessage\n",
    "topic: /navigate_through_poses/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /navigate_to_pose/_action/feedback\ntype: nav2_msgs/action/NavigateToPose_FeedbackMessage\n",
    "topic: /navigate_to_pose/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /odom\ntype: nav_msgs/msg/Odometry\n",
    "topic: /odometry/filtered\ntype: nav_msgs/msg/Odometry\n",
    "topic: /parameter_events\ntype: rcl_interfaces/msg/ParameterEvent\n",
    "topic: /plan\ntype: nav_msgs/msg/Path\n",
    "topic: /plan_smoothed\ntype: nav_msgs/msg/Path\n",
    "topic: /planner_server/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /pose\ntype: geometry_msgs/msg/PoseWithCovarianceStamped\n",
    "topic: /preempt_teleop\ntype: std_msgs/msg/Empty\n",
    "topic: /rosout\ntype: rcl_interfaces/msg/Log\n",
    "topic: /scan\ntype: sensor_msgs/msg/LaserScan\n",
    "topic: /slam_toolbox/feedback\ntype: visualization_msgs/msg/InteractiveMarkerFeedback\n",
    "topic: /slam_toolbox/graph_visualization\ntype: visualization_msgs/msg/MarkerArray\n",
    "topic: /slam_toolbox/scan_visualization\ntype: sensor_msgs/msg/LaserScan\n",
    "topic: /slam_toolbox/update\ntype: visualization_msgs/msg/InteractiveMarkerUpdate\n",
    "topic: /smooth_path/_action/feedback\ntype: nav2_msgs/action/SmoothPath_FeedbackMessage\n",
    "topic: /smooth_path/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /smoother_server/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /speed_limit\ntype: nav2_msgs/msg/SpeedLimit\n",
    "topic: /spin/_action/feedback\ntype: nav2_msgs/action/Spin_FeedbackMessage\n",
    "topic: /spin/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /tf_static\ntype: tf2_msgs/msg/TFMessage\n",
    "topic: /trajectories\ntype: visualization_msgs/msg/MarkerArray\n",
    "topic: /transformed_global_plan\ntype: nav_msgs/msg/Path\n",
    "topic: /unsmoothed_plan\ntype: nav_msgs/msg/Path\n",
    "topic: /velocity_smoother/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
    "topic: /wait/_action/feedback\ntype: nav2_msgs/action/Wait_FeedbackMessage\n",
    "topic: /wait/_action/status\ntype: action_msgs/msg/GoalStatusArray\n",
    "topic: /waypoint_follower/transition_event\ntype: lifecycle_msgs/msg/TransitionEvent\n",
]
ACTIONS_AND_TYPES: Dict[str, str] = {
    "/assisted_teleop": "nav2_msgs/action/AssistedTeleop",
    "/backup": "nav2_msgs/action/BackUp",
    "/compute_path_through_poses": "nav2_msgs/action/ComputePathThroughPoses",
    "/compute_path_to_pose": "nav2_msgs/action/ComputePathToPose",
    "/drive_on_heading": "nav2_msgs/action/DriveOnHeading",
    "/follow_path": "nav2_msgs/action/FollowPath",
    "/follow_waypoints": "nav2_msgs/action/FollowWaypoints",
    "/navigate_through_poses": "nav2_msgs/action/NavigateThroughPoses",
    "/navigate_to_pose": "nav2_msgs/action/NavigateToPose",
    "/smooth_path": "nav2_msgs/action/SmoothPath",
    "/spin": "nav2_msgs/action/Spin",
    "/wait": "nav2_msgs/action/Wait",
}

SERVICES_AND_TYPES: Dict[str, str] = {
    "/assisted_teleop/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/assisted_teleop/_action/get_result": "nav2_msgs/action/AssistedTeleop_GetResult",
    "/assisted_teleop/_action/send_goal": "nav2_msgs/action/AssistedTeleop_SendGoal",
    "/backup/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/backup/_action/get_result": "nav2_msgs/action/BackUp_GetResult",
    "/backup/_action/send_goal": "nav2_msgs/action/BackUp_SendGoal",
    "/behavior_server/change_state": "lifecycle_msgs/srv/ChangeState",
    "/behavior_server/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/behavior_server/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/behavior_server/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/behavior_server/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/behavior_server/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/behavior_server/get_state": "lifecycle_msgs/srv/GetState",
    "/behavior_server/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/behavior_server/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/behavior_server/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/behavior_server/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/bt_navigator/change_state": "lifecycle_msgs/srv/ChangeState",
    "/bt_navigator/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/bt_navigator/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/bt_navigator/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/bt_navigator/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/bt_navigator/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/bt_navigator/get_state": "lifecycle_msgs/srv/GetState",
    "/bt_navigator/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/bt_navigator/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/bt_navigator/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/bt_navigator/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/bt_navigator_navigate_through_poses_rclcpp_node/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/bt_navigator_navigate_through_poses_rclcpp_node/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/bt_navigator_navigate_through_poses_rclcpp_node/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/bt_navigator_navigate_through_poses_rclcpp_node/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/bt_navigator_navigate_through_poses_rclcpp_node/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/bt_navigator_navigate_through_poses_rclcpp_node/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/bt_navigator_navigate_to_pose_rclcpp_node/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/bt_navigator_navigate_to_pose_rclcpp_node/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/bt_navigator_navigate_to_pose_rclcpp_node/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/bt_navigator_navigate_to_pose_rclcpp_node/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/bt_navigator_navigate_to_pose_rclcpp_node/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/bt_navigator_navigate_to_pose_rclcpp_node/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/compute_path_through_poses/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/compute_path_through_poses/_action/get_result": "nav2_msgs/action/ComputePathThroughPoses_GetResult",
    "/compute_path_through_poses/_action/send_goal": "nav2_msgs/action/ComputePathThroughPoses_SendGoal",
    "/compute_path_to_pose/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/compute_path_to_pose/_action/get_result": "nav2_msgs/action/ComputePathToPose_GetResult",
    "/compute_path_to_pose/_action/send_goal": "nav2_msgs/action/ComputePathToPose_SendGoal",
    "/controller_server/change_state": "lifecycle_msgs/srv/ChangeState",
    "/controller_server/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/controller_server/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/controller_server/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/controller_server/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/controller_server/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/controller_server/get_state": "lifecycle_msgs/srv/GetState",
    "/controller_server/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/controller_server/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/controller_server/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/controller_server/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/drive_on_heading/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/drive_on_heading/_action/get_result": "nav2_msgs/action/DriveOnHeading_GetResult",
    "/drive_on_heading/_action/send_goal": "nav2_msgs/action/DriveOnHeading_SendGoal",
    "/follow_path/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/follow_path/_action/get_result": "nav2_msgs/action/FollowPath_GetResult",
    "/follow_path/_action/send_goal": "nav2_msgs/action/FollowPath_SendGoal",
    "/follow_waypoints/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/follow_waypoints/_action/get_result": "nav2_msgs/action/FollowWaypoints_GetResult",
    "/follow_waypoints/_action/send_goal": "nav2_msgs/action/FollowWaypoints_SendGoal",
    "/global_costmap/clear_around_global_costmap": "nav2_msgs/srv/ClearCostmapAroundRobot",
    "/global_costmap/clear_entirely_global_costmap": "nav2_msgs/srv/ClearEntireCostmap",
    "/global_costmap/clear_except_global_costmap": "nav2_msgs/srv/ClearCostmapExceptRegion",
    "/global_costmap/get_costmap": "nav2_msgs/srv/GetCostmap",
    "/global_costmap/global_costmap/change_state": "lifecycle_msgs/srv/ChangeState",
    "/global_costmap/global_costmap/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/global_costmap/global_costmap/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/global_costmap/global_costmap/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/global_costmap/global_costmap/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/global_costmap/global_costmap/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/global_costmap/global_costmap/get_state": "lifecycle_msgs/srv/GetState",
    "/global_costmap/global_costmap/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/global_costmap/global_costmap/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/global_costmap/global_costmap/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/global_costmap/global_costmap/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/grounded_sam/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/grounded_sam/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/grounded_sam/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/grounded_sam/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/grounded_sam/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/grounded_sam/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/grounded_sam_segment": "rai_interfaces/srv/RAIGroundedSam",
    "/grounding_dino/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/grounding_dino/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/grounding_dino/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/grounding_dino/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/grounding_dino/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/grounding_dino/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/grounding_dino_classify": "rai_interfaces/srv/RAIGroundingDino",
    "/is_path_valid": "nav2_msgs/srv/IsPathValid",
    "/launch_ros_138640/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/launch_ros_138640/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/launch_ros_138640/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/launch_ros_138640/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/launch_ros_138640/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/launch_ros_138640/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/lifecycle_manager_navigation/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/lifecycle_manager_navigation/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/lifecycle_manager_navigation/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/lifecycle_manager_navigation/is_active": "std_srvs/srv/Trigger",
    "/lifecycle_manager_navigation/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/lifecycle_manager_navigation/manage_nodes": "nav2_msgs/srv/ManageLifecycleNodes",
    "/lifecycle_manager_navigation/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/lifecycle_manager_navigation/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/lifecycle_manager_slam/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/lifecycle_manager_slam/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/lifecycle_manager_slam/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/lifecycle_manager_slam/is_active": "std_srvs/srv/Trigger",
    "/lifecycle_manager_slam/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/lifecycle_manager_slam/manage_nodes": "nav2_msgs/srv/ManageLifecycleNodes",
    "/lifecycle_manager_slam/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/lifecycle_manager_slam/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/local_costmap/clear_around_local_costmap": "nav2_msgs/srv/ClearCostmapAroundRobot",
    "/local_costmap/clear_entirely_local_costmap": "nav2_msgs/srv/ClearEntireCostmap",
    "/local_costmap/clear_except_local_costmap": "nav2_msgs/srv/ClearCostmapExceptRegion",
    "/local_costmap/get_costmap": "nav2_msgs/srv/GetCostmap",
    "/local_costmap/local_costmap/change_state": "lifecycle_msgs/srv/ChangeState",
    "/local_costmap/local_costmap/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/local_costmap/local_costmap/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/local_costmap/local_costmap/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/local_costmap/local_costmap/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/local_costmap/local_costmap/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/local_costmap/local_costmap/get_state": "lifecycle_msgs/srv/GetState",
    "/local_costmap/local_costmap/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/local_costmap/local_costmap/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/local_costmap/local_costmap/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/local_costmap/local_costmap/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/map_saver/change_state": "lifecycle_msgs/srv/ChangeState",
    "/map_saver/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/map_saver/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/map_saver/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/map_saver/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/map_saver/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/map_saver/get_state": "lifecycle_msgs/srv/GetState",
    "/map_saver/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/map_saver/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/map_saver/save_map": "nav2_msgs/srv/SaveMap",
    "/map_saver/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/map_saver/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/nav2_container/_container/list_nodes": "composition_interfaces/srv/ListNodes",
    "/nav2_container/_container/load_node": "composition_interfaces/srv/LoadNode",
    "/nav2_container/_container/unload_node": "composition_interfaces/srv/UnloadNode",
    "/navigate_through_poses/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/navigate_through_poses/_action/get_result": "nav2_msgs/action/NavigateThroughPoses_GetResult",
    "/navigate_through_poses/_action/send_goal": "nav2_msgs/action/NavigateThroughPoses_SendGoal",
    "/navigate_to_pose/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/navigate_to_pose/_action/get_result": "nav2_msgs/action/NavigateToPose_GetResult",
    "/navigate_to_pose/_action/send_goal": "nav2_msgs/action/NavigateToPose_SendGoal",
    "/o3de_ros2_node/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/o3de_ros2_node/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/o3de_ros2_node/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/o3de_ros2_node/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/o3de_ros2_node/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/o3de_ros2_node/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/planner_server/change_state": "lifecycle_msgs/srv/ChangeState",
    "/planner_server/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/planner_server/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/planner_server/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/planner_server/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/planner_server/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/planner_server/get_state": "lifecycle_msgs/srv/GetState",
    "/planner_server/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/planner_server/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/planner_server/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/planner_server/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/rai_ros2_ari_connector_b6ed00ab6356/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/rai_ros2_ari_connector_b6ed00ab6356/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/rai_ros2_ari_connector_b6ed00ab6356/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/rai_ros2_ari_connector_b6ed00ab6356/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/rai_ros2_ari_connector_b6ed00ab6356/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/rai_ros2_ari_connector_b6ed00ab6356/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/slam_toolbox/clear_changes": "slam_toolbox/srv/Clear",
    "/slam_toolbox/clear_queue": "slam_toolbox/srv/ClearQueue",
    "/slam_toolbox/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/slam_toolbox/deserialize_map": "slam_toolbox/srv/DeserializePoseGraph",
    "/slam_toolbox/dynamic_map": "nav_msgs/srv/GetMap",
    "/slam_toolbox/get_interactive_markers": "visualization_msgs/srv/GetInteractiveMarkers",
    "/slam_toolbox/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/slam_toolbox/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/slam_toolbox/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/slam_toolbox/manual_loop_closure": "slam_toolbox/srv/LoopClosure",
    "/slam_toolbox/pause_new_measurements": "slam_toolbox/srv/Pause",
    "/slam_toolbox/save_map": "slam_toolbox/srv/SaveMap",
    "/slam_toolbox/serialize_map": "slam_toolbox/srv/SerializePoseGraph",
    "/slam_toolbox/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/slam_toolbox/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/slam_toolbox/toggle_interactive_mode": "slam_toolbox/srv/ToggleInteractive",
    "/smooth_path/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/smooth_path/_action/get_result": "nav2_msgs/action/SmoothPath_GetResult",
    "/smooth_path/_action/send_goal": "nav2_msgs/action/SmoothPath_SendGoal",
    "/smoother_server/change_state": "lifecycle_msgs/srv/ChangeState",
    "/smoother_server/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/smoother_server/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/smoother_server/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/smoother_server/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/smoother_server/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/smoother_server/get_state": "lifecycle_msgs/srv/GetState",
    "/smoother_server/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/smoother_server/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/smoother_server/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/smoother_server/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/spin/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/spin/_action/get_result": "nav2_msgs/action/Spin_GetResult",
    "/spin/_action/send_goal": "nav2_msgs/action/Spin_SendGoal",
    "/tf2_frames": "tf2_msgs/srv/FrameGraph",
    "/velocity_smoother/change_state": "lifecycle_msgs/srv/ChangeState",
    "/velocity_smoother/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/velocity_smoother/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/velocity_smoother/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/velocity_smoother/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/velocity_smoother/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/velocity_smoother/get_state": "lifecycle_msgs/srv/GetState",
    "/velocity_smoother/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/velocity_smoother/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/velocity_smoother/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/velocity_smoother/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
    "/wait/_action/cancel_goal": "action_msgs/srv/CancelGoal",
    "/wait/_action/get_result": "nav2_msgs/action/Wait_GetResult",
    "/wait/_action/send_goal": "nav2_msgs/action/Wait_SendGoal",
    "/waypoint_follower/change_state": "lifecycle_msgs/srv/ChangeState",
    "/waypoint_follower/describe_parameters": "rcl_interfaces/srv/DescribeParameters",
    "/waypoint_follower/get_available_states": "lifecycle_msgs/srv/GetAvailableStates",
    "/waypoint_follower/get_available_transitions": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/waypoint_follower/get_parameter_types": "rcl_interfaces/srv/GetParameterTypes",
    "/waypoint_follower/get_parameters": "rcl_interfaces/srv/GetParameters",
    "/waypoint_follower/get_state": "lifecycle_msgs/srv/GetState",
    "/waypoint_follower/get_transition_graph": "lifecycle_msgs/srv/GetAvailableTransitions",
    "/waypoint_follower/list_parameters": "rcl_interfaces/srv/ListParameters",
    "/waypoint_follower/set_parameters": "rcl_interfaces/srv/SetParameters",
    "/waypoint_follower/set_parameters_atomically": "rcl_interfaces/srv/SetParametersAtomically",
}
INTERFACES: Dict[str, str] = {
    "/navigate_to_pose": """#goal definition
geometry_msgs/PoseStamped pose
	std_msgs/Header header
		builtin_interfaces/Time stamp
			int32 sec
			uint32 nanosec
		string frame_id
	Pose pose
		Point position
			float64 x
			float64 y
			float64 z
		Quaternion orientation
			float64 x 0
			float64 y 0
			float64 z 0
			float64 w 1
string behavior_tree
---
#result definition
std_msgs/Empty result
---
#feedback definition
geometry_msgs/PoseStamped current_pose
	std_msgs/Header header
		builtin_interfaces/Time stamp
			int32 sec
			uint32 nanosec
		string frame_id
	Pose pose
		Point position
			float64 x
			float64 y
			float64 z
		Quaternion orientation
			float64 x 0
			float64 y 0
			float64 z 0
			float64 w 1
builtin_interfaces/Duration navigation_time
	int32 sec
	uint32 nanosec
builtin_interfaces/Duration estimated_time_remaining
	int32 sec
	uint32 nanosec
int16 number_of_recoveries
float32 distance_remaining
"""
}

ACTION_MODELS: Dict[str, Type[BaseModel]] = {
    "nav2_msgs/action/NavigateToPose": NavigateToPoseGoal,
    "nav2_msgs/action/Spin": SpinGoal,
}

ACTION_STRINGS = [
    f"action: {action}\ntype: {act_type}\n"
    for action, act_type in ACTIONS_AND_TYPES.items()
]

SERVICE_STRINGS = [
    f"service: {service}\ntype: {srv_type}\n"
    for service, srv_type in SERVICES_AND_TYPES.items()
]


class NavigationTask(Task):
    @property
    def type(self) -> str:
        return "navigation"

    def get_system_prompt(self) -> str:
        return ROBOT_NAVIGATION_SYSTEM_PROMPT

    @property
    def available_tools(self) -> List[BaseTool]:
        tools = MockActionsToolkit(
            mock_actions_names_and_types=ACTION_STRINGS,
            available_actions=list(ACTIONS_AND_TYPES.keys()),
            available_action_types=list(ACTIONS_AND_TYPES.values()),
            available_action_models=ACTION_MODELS,
        ).get_tools()
        tools.append(MockGetROS2MessageInterfaceTool(mock_interfaces=INTERFACES))
        return tools


class NavigateToPointTask(NavigationTask):
    complexity = "medium"

    def get_prompt(self) -> str:
        return "Navigate to the point (2.0, 2.0, 0.0). Remember to list actions and get interface"


class SpinAroundTask(NavigationTask):
    recursion_limit = 50
    complexity = "medium"

    def get_prompt(self) -> str:
        return "Spin around by 3 radians."


class MoveToFrontTask(NavigationTask):
    recursion_limit = 50
    complexity = "medium"

    def get_prompt(self) -> str:
        return "Move 2 meters to the front."


class MoveToBedTask(NavigationTask):
    recursion_limit = 50
    complexity = "medium"

    @property
    def available_tools(self) -> List[BaseTool]:
        return [
            MockGetROS2ActionsNamesAndTypesTool(
                mock_actions_names_and_types=ACTION_STRINGS
            ),
            MockStartROS2ActionTool(
                available_actions=list(ACTIONS_AND_TYPES.keys()),
                available_action_types=list(ACTIONS_AND_TYPES.values()),
                available_action_models=ACTION_MODELS,
            ),
            MockGetROS2ActionFeedbackTool(),
            MockGetROS2ActionResultTool(),
            MockGetROS2MessageInterfaceTool(mock_interfaces=INTERFACES),
            MockGetROS2TopicsNamesAndTypesTool(
                mock_topics_names_and_types=TOPICS_NAMES_AND_TYPES
            ),
            MockGetDistanceToObjectsTool(
                available_topics=[
                    "/camera/camera/color/image_raw",
                    "/camera/camera/depth/image_rect_raw",
                ],
                mock_distance_measurements=[
                    DistanceMeasurement(name="bed", distance=5.0)
                ],
            ),
        ]

    def get_prompt(self) -> str:
        return "Move closer to the to the bed. Leave 1 meter of space between the bed and you."



# Copyright (C) 2025 Robotec.AI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional

from rai.types import PoseStamped, ROS2BaseModel, Time


class TaskGoal(ROS2BaseModel):
    task: Optional[str] = ""
    description: Optional[str] = ""
    priority: Optional[str] = ""


class TaskResult(ROS2BaseModel):
    success: Optional[bool] = False
    report: Optional[str] = ""


class TaskFeedback(ROS2BaseModel):
    current_status: Optional[str] = ""


class LoadMapRequest(ROS2BaseModel):
    filename: Optional[str] = ""


class LoadMapResponse(ROS2BaseModel):
    success: Optional[bool] = False


class NavigateToPoseGoal(ROS2BaseModel):
    pose: Optional[PoseStamped] = None
    behavior_tree: Optional[str] = None


class ActionResult(ROS2BaseModel):
    result: Optional[Dict[str, Any]] = None


class NavigateToPoseFeedback(ROS2BaseModel):
    current_pose: Optional[PoseStamped] = None
    navigation_time: Optional[Time] = None
    estimated_time_remaining: Optional[Time] = None
    number_of_recoveries: Optional[int] = None
    distance_remaining: Optional[float] = None


class NavigateToPoseAction(ROS2BaseModel):
    goal: Optional[NavigateToPoseGoal] = None
    result: Optional[ActionResult] = None
    feedback: Optional[NavigateToPoseFeedback] = None


class SpinGoal(ROS2BaseModel):
    target_yaw: Optional[float] = None
    time_allowance: Optional[Time] = None


class SpinFeedback(ROS2BaseModel):
    angle_turned: Optional[float] = None
    remaining_yaw: Optional[float] = None


class SpinAction(ROS2BaseModel):
    goal: Optional[SpinGoal] = None
    result: Optional[ActionResult] = None
    feedback: Optional[SpinFeedback] = None
